import time

import time

import pytest
from indy.ledger import build_get_revoc_reg_def_request, build_get_revoc_reg_request, build_get_revoc_reg_delta_request

from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.constants import GET_ATTR, GET_NYM, SCHEMA, GET_SCHEMA, \
    CLAIM_DEF, REVOCATION, GET_CLAIM_DEF, CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_FROM, \
    SCHEMA_ATTR_NAMES, SCHEMA_NAME, SCHEMA_VERSION, CLAIM_DEF_TAG, ENDORSER, JSON_LD_CONTEXT, RICH_SCHEMA, \
    RICH_SCHEMA_ENCODING, RICH_SCHEMA_MAPPING, RICH_SCHEMA_CRED_DEF, RS_CONTEXT_TYPE_VALUE, \
    RS_SCHEMA_TYPE_VALUE, RS_ENCODING_TYPE_VALUE, RS_MAPPING_TYPE_VALUE, RS_CRED_DEF_TYPE_VALUE, \
    GET_RICH_SCHEMA_OBJECT_BY_ID, RS_ID, GET_RICH_SCHEMA_OBJECT_BY_METADATA, RS_NAME, RS_VERSION, RS_TYPE, \
    RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE
from indy_common.serialization import attrib_raw_data_serializer
from indy_node.test.anon_creds.helper import get_revoc_reg_def_id
from indy_node.test.api.helper import sdk_write_rich_schema_object_and_check
from indy_node.test.auth_rule.helper import sdk_send_and_check_get_auth_rule_request, generate_key
from indy_node.test.rich_schema.templates import W3C_BASE_CONTEXT, RICH_SCHEMA_EX1, RICH_SCHEMA_ENCODING_EX1, \
    RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_CRED_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1
from indy_node.test.state_proof.helper import check_valid_proof, \
    sdk_submit_operation_and_get_result
from plenum.common.constants import TARGET_NYM, TXN_TYPE, RAW, DATA, \
    ROLE, VERKEY, TXN_TIME, NYM, NAME, VERSION
from plenum.common.types import f

# Fixtures, do not remove
from indy_node.test.attrib_txn.test_nym_attrib import sdk_added_raw_attribute, attributeName, attributeValue, \
    attributeData
from indy_node.test.schema.test_send_get_schema import send_schema_seq_no
from indy_node.test.anon_creds.conftest import send_revoc_reg_entry, send_revoc_reg_def, send_claim_def, claim_def
from indy_node.test.schema.test_send_get_schema import send_schema_req
from plenum.common.util import randomString, SortedDict

from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_req


def test_state_proof_returned_for_get_attr(looper,
                                           nodeSetWithOneNodeResponding,
                                           sdk_added_raw_attribute,
                                           attributeName,
                                           attributeData,
                                           sdk_pool_handle,
                                           sdk_wallet_client):
    """
    Tests that state proof is returned in the reply for GET_ATTR transactions.
    Use different submitter and reader!
    """
    get_attr_operation = {
        TARGET_NYM: sdk_added_raw_attribute['result']['txn']['data']['dest'],
        TXN_TYPE: GET_ATTR,
        RAW: attributeName
    }

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_attr_operation)
    expected_data = attrib_raw_data_serializer.deserialize(attributeData)
    assert DATA in result
    data = attrib_raw_data_serializer.deserialize(result[DATA])
    assert data == expected_data
    assert result[TXN_TIME]
    check_valid_proof(result)


def test_state_proof_returned_for_get_nym(looper,
                                          nodeSetWithOneNodeResponding,
                                          sdk_user_wallet_a,
                                          sdk_pool_handle,
                                          sdk_wallet_client,
                                          sdk_wallet_endorser):
    """
    Tests that state proof is returned in the reply for GET_NYM transactions.
    Use different submitter and reader!
    """
    _, dest = sdk_user_wallet_a

    nym_operation = {
        TARGET_NYM: dest,
        TXN_TYPE: NYM
    }

    sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                        sdk_wallet_endorser,
                                        nym_operation)

    get_nym_operation = {
        TARGET_NYM: dest,
        TXN_TYPE: GET_NYM
    }

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_nym_operation)
    assert DATA in result
    assert result[DATA]
    data = domain_state_serializer.deserialize(result[DATA])
    assert ROLE in data
    assert VERKEY in data
    assert f.IDENTIFIER.nm in data
    assert result[TXN_TIME]
    check_valid_proof(result)


def test_state_proof_returned_for_get_schema(looper,
                                             nodeSetWithOneNodeResponding,
                                             sdk_wallet_endorser,
                                             sdk_pool_handle,
                                             sdk_wallet_client):
    """
    Tests that state proof is returned in the reply for GET_SCHEMA transactions.
    Use different submitter and reader!
    """
    _, dest = sdk_wallet_endorser
    schema_name = "test_schema"
    schema_version = "1.0"
    schema_attr_names = ["width", "height"]
    data = {
        SCHEMA_NAME: schema_name,
        SCHEMA_VERSION: schema_version,
        SCHEMA_ATTR_NAMES: schema_attr_names
    }
    schema_operation = {
        TXN_TYPE: SCHEMA,
        DATA: data
    }
    sdk_submit_operation_and_get_result(looper,
                                        sdk_pool_handle,
                                        sdk_wallet_endorser,
                                        schema_operation)

    get_schema_operation = {
        TARGET_NYM: dest,
        TXN_TYPE: GET_SCHEMA,
        DATA: {
            NAME: schema_name,
            VERSION: schema_version,
        }
    }
    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_schema_operation)
    assert DATA in result
    data = result.get(DATA)
    assert data
    assert SCHEMA_ATTR_NAMES in data
    assert data[SCHEMA_ATTR_NAMES] == schema_attr_names
    assert NAME in data
    assert VERSION in data
    assert result[TXN_TIME]
    check_valid_proof(result)


# The order of creation is essential as some rich schema object reference others by ID
# Encoding's id must be equal to the one used in RICH_SCHEMA_MAPPING_EX1

# txn_type, rs_type, content, rs_id, rs_name, rs_version:
PARAMS = [
    (
        JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, W3C_BASE_CONTEXT, randomString(), randomString(), '1.0'
    ),
    (
        RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, RICH_SCHEMA_EX1, RICH_SCHEMA_EX1['@id'], randomString(), '2.0'
    ),
    (
        RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA_ENCODING_EX1,
        "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD", randomString(), '1.1'
    ),
    (
        RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_MAPPING_EX1['@id'],
        randomString(),
        '10.0'
    ),
    (
        RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA_CRED_DEF_EX1, randomString(),
        randomString(), '1000.0'
    ),
    (
        RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE, RICH_SCHEMA_PRES_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1['@id'],
        randomString(), '1.1000.1'
    )
]


@pytest.fixture(scope='module')
def write_rich_schema(looper, sdk_pool_handle, sdk_wallet_endorser):
    for txn_type, rs_type, content, rs_id, rs_name, rs_version in PARAMS:
        sdk_write_rich_schema_object_and_check(looper, sdk_wallet_endorser, sdk_pool_handle,
                                               txn_type=txn_type, rs_id=rs_id, rs_name=rs_name,
                                               rs_version=rs_version, rs_type=rs_type, rs_content=content)


@pytest.mark.skip
# TODO fix this test so it does not rely on Indy-SDK,
# or, fix this test once Rich Schema requests are part of Indy-SDK
@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_state_proof_returned_for_get_rich_schema(looper,
                                                  nodeSetWithOneNodeResponding,
                                                  sdk_wallet_endorser,
                                                  sdk_pool_handle,
                                                  sdk_wallet_client,
                                                  write_rich_schema,
                                                  txn_type, rs_type, content, rs_id, rs_name, rs_version):
    """
    Tests that state proof is returned in the reply for GET_RICH_SCHEMA_OBJECT_BY_ID.
    Use different submitter and reader!
    """
    get_rich_schema_by_id_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_ID,
        RS_ID: rs_id,
    }
    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_rich_schema_by_id_operation)
    expected_data = SortedDict({
        'id': rs_id,
        'rsType': rs_type,
        'rsName': rs_name,
        'rsVersion': rs_version,
        'content': content,
        'from': sdk_wallet_endorser[1]
    })
    assert SortedDict(result['data']) == expected_data
    assert result['seqNo']
    assert result['txnTime']
    assert result['state_proof']
    check_valid_proof(result)


@pytest.mark.skip
# TODO fix this test so it does not rely on Indy-SDK,
# or, fix this test once Rich Schema objects are part of Indy-SDK
@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_state_proof_returned_for_get_rich_schema_obj_by_metadata(looper,
                                                                  nodeSetWithOneNodeResponding,
                                                                  sdk_wallet_endorser,
                                                                  sdk_pool_handle,
                                                                  sdk_wallet_client,
                                                                  write_rich_schema,
                                                                  txn_type, rs_type, content, rs_id, rs_name,
                                                                  rs_version):
    """
    Tests that state proof is returned in the reply for GET_RICH_SCHEMA_OBJECT_BY_METADATA.
    Use different submitter and reader!
    """

    get_rich_schema_by_metadata_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_METADATA,
        RS_NAME: rs_name,
        RS_VERSION: rs_version,
        RS_TYPE: rs_type
    }

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_rich_schema_by_metadata_operation)
    expected_data = SortedDict({
        'id': rs_id,
        'rsType': rs_type,
        'rsName': rs_name,
        'rsVersion': rs_version,
        'content': content,
        'from': sdk_wallet_endorser[1]
    })
    assert SortedDict(result['data']) == expected_data
    assert result['seqNo']
    assert result['txnTime']
    assert result['state_proof']
    check_valid_proof(result)


def test_state_proof_returned_for_get_claim_def(looper,
                                                nodeSetWithOneNodeResponding,
                                                sdk_wallet_endorser,
                                                sdk_pool_handle,
                                                sdk_wallet_client,
                                                send_schema_seq_no):
    """
    Tests that state proof is returned in the reply for GET_CLAIM_DEF
    transactions.
    Use different submitter and reader!
    """
    _, dest = sdk_wallet_endorser
    data = {"primary": {'N': '123'}, REVOCATION: {'h0': '456'}}
    claim_def_operation = {
        TXN_TYPE: CLAIM_DEF,
        CLAIM_DEF_SCHEMA_REF: send_schema_seq_no,
        DATA: data,
        CLAIM_DEF_SIGNATURE_TYPE: 'CL',
        CLAIM_DEF_TAG: "tag1"
    }
    sdk_submit_operation_and_get_result(looper,
                                        sdk_pool_handle,
                                        sdk_wallet_endorser,
                                        claim_def_operation)
    get_claim_def_operation = {
        CLAIM_DEF_FROM: dest,
        TXN_TYPE: GET_CLAIM_DEF,
        CLAIM_DEF_SCHEMA_REF: send_schema_seq_no,
        CLAIM_DEF_SIGNATURE_TYPE: 'CL',
        CLAIM_DEF_TAG: "tag1"
    }
    result = sdk_submit_operation_and_get_result(looper,
                                                 sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_claim_def_operation)
    expected_data = data
    assert DATA in result
    data = result.get(DATA)
    assert data
    assert data == expected_data
    assert result[TXN_TIME]
    check_valid_proof(result)


def test_state_proof_returned_for_get_auth_rule(looper,
                                                nodeSetWithOneNodeResponding,
                                                sdk_wallet_steward,
                                                sdk_pool_handle,
                                                sdk_wallet_client,
                                                send_auth_rule):
    req = send_auth_rule

    key = generate_key(auth_action=ADD_PREFIX, auth_type=NYM,
                       field=ROLE, new_value=ENDORSER)
    rep = sdk_send_and_check_get_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_client, **key)
    result = rep[0][1]['result']

    expected_data = req[0][0]['operation']
    del expected_data['type']
    assert DATA in result
    data = result.get(DATA)
    assert data
    assert data[0] == expected_data
    check_valid_proof(result)


def test_state_proof_returned_for_get_revoc_reg_def(looper,
                                                    nodeSetWithOneNodeResponding,
                                                    sdk_wallet_steward,
                                                    sdk_pool_handle,
                                                    sdk_wallet_client,
                                                    send_revoc_reg_def):
    revoc_reg_def_data = send_revoc_reg_def[0]['operation']

    req = looper.loop.run_until_complete(build_get_revoc_reg_def_request(
        sdk_wallet_client[1], get_revoc_reg_def_id(sdk_wallet_steward[1], send_revoc_reg_def[0])))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, req)])
    result = rep[0][1]['result']

    expected_data = revoc_reg_def_data
    del expected_data['type']
    assert DATA in result
    data = result.get(DATA)
    assert data
    assert data == expected_data
    assert result[TXN_TIME]
    check_valid_proof(result)


def test_state_proof_returned_for_get_revoc_reg_entry(looper,
                                                      nodeSetWithOneNodeResponding,
                                                      sdk_wallet_steward,
                                                      sdk_pool_handle,
                                                      sdk_wallet_client,
                                                      send_revoc_reg_entry):
    revoc_reg_def = send_revoc_reg_entry[0]
    revoc_reg_entry_data = send_revoc_reg_entry[1][0]['operation']

    timestamp = int(time.time())
    req = looper.loop.run_until_complete(build_get_revoc_reg_request(
        sdk_wallet_client[1], get_revoc_reg_def_id(sdk_wallet_steward[1], revoc_reg_def[0]), timestamp))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, req)])
    result = rep[0][1]['result']

    expected_data = revoc_reg_entry_data
    data = result.get(DATA)

    del expected_data['type']
    del data['seqNo']
    del data['txnTime']
    assert DATA in result
    assert data
    assert data == expected_data
    assert result[TXN_TIME]
    check_valid_proof(result)


def check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, timestamp_fr, timestamp_to,
                    sdk_pool_handle, revoc_reg_entry_data, check_data=True):
    req = looper.loop.run_until_complete(build_get_revoc_reg_delta_request(
        sdk_wallet_client[1], get_revoc_reg_def_id(sdk_wallet_steward[1], revoc_reg_def[0]), timestamp_fr,
        timestamp_to))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, req)])

    if check_data:
        result = rep[0][1]['result']
        expected_data = revoc_reg_entry_data
        data = result.get(DATA)['value']['accum_to']

        del data['seqNo']
        del data['txnTime']
        if 'type' in expected_data:
            del expected_data['type']
        assert DATA in result
        assert data
        assert data == expected_data
        assert result[TXN_TIME]
        check_valid_proof(result)


def test_state_proof_returned_for_get_revoc_reg_delta(looper,
                                                      nodeSetWithOneNodeResponding,
                                                      sdk_wallet_steward,
                                                      sdk_pool_handle,
                                                      sdk_wallet_client,
                                                      send_revoc_reg_entry):
    revoc_reg_def = send_revoc_reg_entry[0]
    revoc_reg_entry_data = send_revoc_reg_entry[1][0]['operation']
    timestamp = send_revoc_reg_entry[1][1]['result']['txnMetadata']['txnTime']

    check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, None, timestamp + 1,
                    sdk_pool_handle, revoc_reg_entry_data)

    check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, None, timestamp - 1,
                    sdk_pool_handle, revoc_reg_entry_data, False)

    # TODO: INDY-2115
    # check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, timestamp - 2, timestamp - 1,
    #                 sdk_pool_handle, revoc_reg_entry_data, False)

    # check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, timestamp - 1, timestamp + 1,
    #                 sdk_pool_handle, revoc_reg_entry_data)

    check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, timestamp + 1, timestamp + 2,
                    sdk_pool_handle, revoc_reg_entry_data)

    # TODO: INDY-2115
    # check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, None, timestamp - 999,
    #                 sdk_pool_handle, revoc_reg_entry_data)
    #
    # check_get_delta(looper, sdk_wallet_client, sdk_wallet_steward, revoc_reg_def, timestamp - 1000, timestamp - 999,
    #                 sdk_pool_handle, revoc_reg_entry_data)



import copy
import json
from functools import reduce
from operator import getitem

import pytest

from indy_common.constants import RS_CONTENT, ENDORSER, RS_MAPPING_SCHEMA, RS_ID, \
    RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA, \
    RS_SCHEMA_TYPE_VALUE, RS_MAPPING_ENC, RS_MAPPING_RANK, RS_MAPPING_ATTRIBUTES
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_encoding_handler import \
    RichSchemaEncodingHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_handler import RichSchemaHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_mapping_handler import \
    RichSchemaMappingHandler
from indy_node.test.request_handlers.helper import add_to_idr
from indy_node.test.request_handlers.rich_schema.helper import make_rich_schema_object_exist, rs_req
from indy_node.test.rich_schema.templates import RICH_SCHEMA_ENCODING_EX1, RICH_SCHEMA_EX1
from plenum.common.constants import TRUSTEE
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.util import randomString

TEST_MAPPING = {
    '@id': "did:sov:8a9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
    '@context': "did:sov:2f9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
    '@type': "rdfs:Class",
    "schema": "did:sov:4e9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
    "attributes": {
        "attr1": [
            {
                "enc": "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                "rank": 3
            }
        ],
        "attr2": [
            {
                "enc": "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                "rank": 2
            },
            {
                "enc": "did:sov:2x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                "rank": 1
            },
        ],
        "attr3": {
            "attr4": [
                {
                    "attr5": [
                        {
                            "enc": "did:sov:3x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                            "rank": 4
                        }
                    ]
                },
                {
                    "attr6": [
                        {
                            "enc": "did:sov:3x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                            "rank": 5
                        }
                    ]
                },
            ]
        },
        "issuer": [
            {
                "enc": "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                "rank": 7
            }
        ],
        "issuanceDate": [
            {
                "enc": "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                "rank": 6
            }
        ],
    }
}

TEST_ENCODING_1 = rs_req(RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE,
                         content=RICH_SCHEMA_ENCODING_EX1, id="did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD")

TEST_ENCODING_2 = rs_req(RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE,
                         content=RICH_SCHEMA_ENCODING_EX1, id="did:sov:2x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD")

TEST_ENCODING_3 = rs_req(RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE,
                         content=RICH_SCHEMA_ENCODING_EX1, id="did:sov:3x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD")


@pytest.fixture()
def mapping_handler(db_manager, write_auth_req_validator):
    return RichSchemaMappingHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def rich_schema_handler(db_manager, write_auth_req_validator):
    return RichSchemaHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def encoding_handler(db_manager, write_auth_req_validator):
    return RichSchemaEncodingHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def rich_schema_req():
    id = "did:sov:4e9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD"
    content = copy.deepcopy(RICH_SCHEMA_EX1)
    content['@id'] = id
    return rs_req(RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE,
                  content=content, id=id)


@pytest.fixture()
def mapping_req(rich_schema_handler, encoding_handler, rich_schema_req):
    make_rich_schema_object_exist(rich_schema_handler, rich_schema_req)
    make_rich_schema_object_exist(encoding_handler, TEST_ENCODING_1)
    make_rich_schema_object_exist(encoding_handler, TEST_ENCODING_2)
    make_rich_schema_object_exist(encoding_handler, TEST_ENCODING_3)

    id = randomString()
    content = copy.deepcopy(TEST_MAPPING)
    content['@id'] = id
    req = rs_req(RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE,
                 content=content, id=id)

    add_to_idr(rich_schema_handler.database_manager.idr_cache, req.identifier, TRUSTEE)
    add_to_idr(rich_schema_handler.database_manager.idr_cache, req.endorser, ENDORSER)

    return req


def test_static_validation_pass(mapping_handler, mapping_req):
    mapping_handler.static_validation(mapping_req)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
@pytest.mark.parametrize('missing_field', ['schema', 'attributes'])
def test_static_validation_fail_no_schema_or_attribute(mapping_handler, mapping_req, status, missing_field):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    if status == 'missing':
        content.pop(missing_field, None)
    elif status == 'empty':
        content[missing_field] = ""
    elif status == 'none':
        content[missing_field] = None
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'{}' must be set in 'content'".format(missing_field)):
        mapping_handler.static_validation(mapping_req)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_fail_no_schema_and_attributes(mapping_handler, mapping_req, status):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    if status == 'missing':
        content.pop('attributes', None)
        content.pop('schema', None)
    elif status == 'empty':
        content['attributes'] = {}
        content['schema'] = {}
    elif status == 'none':
        content['attributes'] = None
        content['schema'] = None
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'schema' and 'attributes' must be set in 'content'"):
        mapping_handler.static_validation(mapping_req)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
@pytest.mark.parametrize('missing_field', ['issuer', 'issuanceDate'])
def test_static_validation_fail_no_issuer_or_issuance_date(mapping_handler, mapping_req, status, missing_field):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    if status == 'missing':
        content[RS_MAPPING_ATTRIBUTES].pop(missing_field, None)
    elif status == 'empty':
        content[RS_MAPPING_ATTRIBUTES][missing_field] = {}
    elif status == 'none':
        content[RS_MAPPING_ATTRIBUTES][missing_field] = None
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'{}' must be in content's 'attributes'".format(missing_field)):
        mapping_handler.static_validation(mapping_req)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_fail_no_issuance_date_and_issuer(mapping_handler, mapping_req, status):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    if status == 'missing':
        content[RS_MAPPING_ATTRIBUTES].pop('issuanceDate', None)
        content[RS_MAPPING_ATTRIBUTES].pop('issuer', None)
    elif status == 'empty':
        content[RS_MAPPING_ATTRIBUTES]['issuanceDate'] = {}
        content[RS_MAPPING_ATTRIBUTES]['issuer'] = {}
    elif status == 'none':
        content[RS_MAPPING_ATTRIBUTES]['issuanceDate'] = None
        content[RS_MAPPING_ATTRIBUTES]['issuer'] = None
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'issuer' and 'issuanceDate' must be in content's 'attributes'"):
        mapping_handler.static_validation(mapping_req)


def test_schema_dynamic_validation_passes(mapping_handler, mapping_req):
    mapping_handler.dynamic_validation(mapping_req, 0)


def test_dynamic_validation_not_existent_schema(mapping_handler, mapping_req):
    schema_id = randomString()
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    content[RS_MAPPING_SCHEMA] = schema_id
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match='Can not find a schema with id={}; please make sure that it has been added to the ledger'.format(
                           schema_id)):
        mapping_handler.dynamic_validation(mapping_req, 0)


def test_dynamic_validation_not_schema_in_schema_field(mapping_handler, mapping_req):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    content[RS_MAPPING_SCHEMA] = TEST_ENCODING_1.operation[RS_ID]
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'schema' field must reference a schema with rsType=sch"):
        mapping_handler.dynamic_validation(mapping_req, 0)


def get_mapping_attr_value(keys, mapping_content):
    return reduce(getitem, keys, mapping_content)


# a test against TEST_MAPPING
@pytest.mark.parametrize('enc_path, index', [
    (['attr1'], 0),
    (['attr2'], 0),
    (['attr2'], 1),
    (['attr3', 'attr4', 0, 'attr5'], 0),
    (['attr3', 'attr4', 1, 'attr6'], 0)
])
@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
@pytest.mark.parametrize('missing_field', [RS_MAPPING_ENC, RS_MAPPING_RANK])
def test_dynamic_validation_empty_field_in_encoding_desc(mapping_handler, mapping_req,
                                                         enc_path, index, status, missing_field):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    enc_dict = get_mapping_attr_value(enc_path, content[RS_MAPPING_ATTRIBUTES])[index]
    if status == 'missing':
        enc_dict.pop(missing_field, None)
    elif status == 'empty':
        enc_dict[missing_field] = ""
    elif status == 'none':
        enc_dict[missing_field] = None
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="{} must be set for the attribute '{}'".format(missing_field, enc_path[-1])):
        mapping_handler.dynamic_validation(mapping_req, 0)


# a test against TEST_MAPPING
@pytest.mark.parametrize('enc_path, index', [
    (['attr1'], 0),
    (['attr2'], 0),
    (['attr2'], 1),
    (['attr3', 'attr4', 0, 'attr5'], 0),
    (['attr3', 'attr4', 1, 'attr6'], 0)
])
@pytest.mark.parametrize('status', ['missing', 'empty', 'none', 'not_a_dict'])
def test_dynamic_validation_empty_encoding_desc(mapping_handler, mapping_req,
                                                enc_path, index, status):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    enc_list = get_mapping_attr_value(enc_path, content[RS_MAPPING_ATTRIBUTES])
    enc_dict = enc_list[index]
    if status == 'missing':
        enc_dict.pop(RS_MAPPING_ENC, None)
        enc_dict.pop(RS_MAPPING_RANK, None)
    elif status == 'empty':
        enc_dict[RS_MAPPING_ENC] = ""
        enc_dict[RS_MAPPING_RANK] = ""
    elif status == 'none':
        enc_dict[RS_MAPPING_ENC] = None
        enc_dict[RS_MAPPING_RANK] = None
    elif status == 'not_a_dict':
        enc_list[index] = "aaaa"
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="enc and rank must be set for the attribute '{}'".format(enc_path[-1])):
        mapping_handler.dynamic_validation(mapping_req, 0)


# a test against TEST_MAPPING
@pytest.mark.parametrize('enc_path, index', [
    (['attr1'], 0),
    (['attr2'], 0),
    (['attr2'], 1),
    (['attr3', 'attr4', 0, 'attr5'], 0),
    (['attr3', 'attr4', 1, 'attr6'], 0)
])
def test_dynamic_validation_not_existent_encoding(mapping_handler, mapping_req,
                                                  enc_path, index):
    wrong_id = randomString()
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    enc_dict = get_mapping_attr_value(enc_path, content[RS_MAPPING_ATTRIBUTES])[index]
    enc_dict[RS_MAPPING_ENC] = wrong_id
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="Can not find a referenced 'enc' with id={} in the '{}' attribute; please make sure that it has been added to the ledger".format(
                           wrong_id, enc_path[-1])):
        mapping_handler.dynamic_validation(mapping_req, 0)


# a test against TEST_MAPPING
@pytest.mark.parametrize('enc_path, index', [
    (['attr1'], 0),
    (['attr2'], 0),
    (['attr2'], 1),
    (['attr3', 'attr4', 0, 'attr5'], 0),
    (['attr3', 'attr4', 1, 'attr6'], 0)
])
def test_dynamic_validation_not_encoding_in_enc_field(mapping_handler, mapping_req,
                                                      rich_schema_req,
                                                      enc_path, index):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    enc_dict = get_mapping_attr_value(enc_path, content[RS_MAPPING_ATTRIBUTES])[index]
    enc_dict[RS_MAPPING_ENC] = rich_schema_req.operation[RS_ID]
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'enc' field in the '{}' attribute must reference an encoding with rsType=enc".format(
                           enc_path[-1])):
        mapping_handler.dynamic_validation(mapping_req, 0)


# a test against TEST_MAPPING
@pytest.mark.parametrize('enc_path, index', [
    (['attr1'], 0),
    (['attr2'], 0),
    (['attr2'], 1),
    (['attr3', 'attr4', 0, 'attr5'], 0),
    (['attr3', 'attr4', 1, 'attr6'], 0),
    (['issuer'], 0),
    (['issuanceDate'], 0)
])
@pytest.mark.parametrize('rank_value', [0, 8, 9, 12, -1])
def test_dynamic_validation_rank_sequence(mapping_handler, mapping_req,
                                          rich_schema_req,
                                          enc_path, index, rank_value):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    enc_dict = get_mapping_attr_value(enc_path, content[RS_MAPPING_ATTRIBUTES])[index]
    enc_dict[RS_MAPPING_RANK] = rank_value
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="the attribute's ranks are not sequential: expected ranks are all values from 1 to 7"):
        mapping_handler.dynamic_validation(mapping_req, 0)


@pytest.mark.parametrize('enc_path, index', [
    (['attr1'], 0),
    (['attr2'], 0),
    (['attr2'], 1),
])
@pytest.mark.parametrize('rank_value', [4, 5, 6, 7])
def test_dynamic_validation_rank_same_rank(mapping_handler, mapping_req,
                                           rich_schema_req,
                                           enc_path, index, rank_value):
    content = copy.deepcopy(json.loads(mapping_req.operation[RS_CONTENT]))
    enc_dict = get_mapping_attr_value(enc_path, content[RS_MAPPING_ATTRIBUTES])[index]
    enc_dict[RS_MAPPING_RANK] = rank_value
    mapping_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="the attribute's ranks are not sequential: expected ranks are all values from 1 to 7"):
        mapping_handler.dynamic_validation(mapping_req, 0)



import json
import logging
import types
from collections import OrderedDict
from typing import Any

import sys

import pytest
from stp_core.common.log import getlogger

from plenum.common.constants import STEWARD_STRING
from plenum.common.util import randomString
from plenum.common.messages.node_messages import Commit

from plenum.server.node import Node
from plenum.test.pool_transactions.helper import sdk_add_new_nym, prepare_nym_request, \
    sdk_sign_and_send_prepared_request
from plenum.test.helper import sdk_json_to_request_object
from pympler import asizeof

max_depth = 10


# Self made memory function. We can use it if we want to explore
# something specific.
def get_max(obj, seen=None, now_depth=0, path=str()):
    if now_depth > max_depth:
        return {}
    dictionary = {(path, type(obj)): sys.getsizeof(obj)}
    path += str(type(obj)) + ' ---> '
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return {}
    seen.add(obj_id)
    if isinstance(obj, dict):
        vpath = path + 'value ---> '
        for d in [get_max(v, seen, now_depth + 1, vpath) for v in obj.values()]:
            updater(dictionary, d)
        kpath = path + 'key ---> '
        for d in [get_max(k, seen, now_depth + 1, kpath) for k in obj.keys()]:
            updater(dictionary, d)
    elif hasattr(obj, '__dict__'):
        dpath = path + '__dict__ ---> '
        d = get_max(obj.__dict__, seen, now_depth + 1, dpath)
        updater(dictionary, d)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        ipath = path + '__iter__ ---> '
        for d in [get_max(i, seen, now_depth + 1, ipath) for i in obj]:
            updater(dictionary, d)
    return dictionary


def updater(store_d, new_d):
    for k in new_d.keys():
        if k in store_d:
            store_d[k] += int(new_d[k])
        else:
            store_d[k] = new_d[k]


def dont_send_commit(self, msg: Any, *rids, signer=None, message_splitter=None):
    if isinstance(msg, (Commit)):
        if rids:
            rids = [rid for rid in rids if rid not in self.nodestack.getRemote(self.ignore_node_name).uid]
        else:
            rids = [self.nodestack.getRemote(name).uid for name
                    in self.nodestack.remotes.keys() if name not in self.ignore_node_name]
    self.old_send(msg, *rids, signer=signer, message_splitter=message_splitter)


def dont_send_commit_to(nodes, ignore_node_name):
    for node in nodes:
        if not hasattr(node, 'ignore_node_name'):
            node.ignore_node_name = []
        node.ignore_node_name.append(ignore_node_name)
        node.old_send = types.MethodType(Node.send, node)
        node.send = types.MethodType(dont_send_commit, node)


def reset_sending(nodes):
    for node in nodes:
        node.send = types.MethodType(Node.send, node)


def sdk_add_new_nym_without_waiting(looper, sdk_pool_handle, creators_wallet,
                                    alias=None, role=None, seed=None,
                                    dest=None, verkey=None, skipverkey=False):
    seed = seed or randomString(32)
    alias = alias or randomString(5)
    wh, _ = creators_wallet

    nym_request, new_did = looper.loop.run_until_complete(
        prepare_nym_request(creators_wallet, seed,
                            alias, role, dest, verkey, skipverkey))
    sdk_sign_and_send_prepared_request(looper, creators_wallet,
                                       sdk_pool_handle, nym_request)


# Pytest logger is heavy, so we exclude it
@pytest.fixture
def logger():
    logger = getlogger()
    old_value = logger.getEffectiveLevel()
    logger.root.setLevel(logging.CRITICAL)
    yield logger
    logger.root.setLevel(old_value)


@pytest.mark.skip('Unskip if you need to debug')
def test_memory_debugging(looper,
                          nodeSet,
                          sdk_wallet_endorser,
                          sdk_pool_handle,
                          logger):
    # Settings
    requests_count = 500
    file_name = '.memory_data.txt'

    # Sets for emulating commits problems
    set1 = list(nodeSet)
    set1.remove(nodeSet[0])
    set2 = list(nodeSet)
    set2.remove(nodeSet[1])
    set3 = list(nodeSet)
    set3.remove(nodeSet[2])
    primary = nodeSet[0]

    memory_dicts = OrderedDict()

    memory_dicts['After starting'] = asizeof.asized(primary, detail=15)

    while primary.master_replica.lastPrePrepareSeqNo < requests_count:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_endorser)

    memory_dicts['After ordering'] = asizeof.asized(primary, detail=15)

    # Emulate commit sending problems
    dont_send_commit_to(set1, nodeSet[0].name)
    dont_send_commit_to(set2, nodeSet[1].name)
    dont_send_commit_to(set3, nodeSet[2].name)

    # Sending requests until nodes generate `unordered_requests_count` 3pc batches
    while primary.master_replica.lastPrePrepareSeqNo < requests_count * 2:
        sdk_add_new_nym_without_waiting(looper, sdk_pool_handle, sdk_wallet_endorser)

    memory_dicts['After {} unordered'.format(requests_count)] = asizeof.asized(primary, detail=15)

    # Remove commit problems
    reset_sending(set1)
    reset_sending(set2)
    reset_sending(set3)

    # primary ask for commits
    for i in range(primary.master_replica.last_ordered_3pc[1], primary.master_replica.lastPrePrepareSeqNo):
        primary.replicas._replicas.values()[0]._request_commit((0, i))
    for i in range(primary.replicas._replicas.values()[1].last_ordered_3pc[1],
                   primary.replicas._replicas.values()[1].lastPrePrepareSeqNo):
        primary.replicas._replicas.values()[1]._request_commit((0, i))
    looper.runFor(5)

    memory_dicts['After {} ordered'.format(requests_count)] = asizeof.asized(primary, detail=15)

    # primary clear queues
    primary.replicas._replicas.values()[0]._gc(primary.replicas._replicas.values()[0].last_ordered_3pc)
    primary.replicas._replicas.values()[1]._gc(primary.replicas._replicas.values()[1].last_ordered_3pc)

    memory_dicts['After _gc called'] = asizeof.asized(primary, detail=15)

    # Emulate problems again
    dont_send_commit_to(set1, nodeSet[0].name)
    dont_send_commit_to(set2, nodeSet[1].name)
    dont_send_commit_to(set3, nodeSet[2].name)

    while primary.master_replica.lastPrePrepareSeqNo < requests_count * 3:
        sdk_add_new_nym_without_waiting(looper, sdk_pool_handle, sdk_wallet_endorser)

    memory_dicts['After {} unordered again'.format(requests_count)] = asizeof.asized(primary, detail=15)

    # Remove commit problems
    reset_sending(set1)
    reset_sending(set2)
    reset_sending(set3)

    for i in range(primary.master_replica.last_ordered_3pc[1], primary.master_replica.lastPrePrepareSeqNo):
        primary.replicas._replicas.values()[0]._request_commit((0, i))
    for i in range(primary.replicas._replicas.values()[1].last_ordered_3pc[1],
                   primary.replicas._replicas.values()[1].lastPrePrepareSeqNo):
        primary.replicas._replicas.values()[1]._request_commit((0, i))
    looper.runFor(5)

    memory_dicts['After {} ordered again'.format(requests_count)] = asizeof.asized(primary, detail=15)

    primary.replicas._replicas.values()[0]._gc(primary.replicas._replicas.values()[0].last_ordered_3pc)
    primary.replicas._replicas.values()[1]._gc(primary.replicas._replicas.values()[1].last_ordered_3pc)

    memory_dicts['After _gc called again'] = asizeof.asized(primary, detail=15)

    file = open(file_name, 'w')
    indent = 75
    for k, size_obj in memory_dicts.items():
        # Formatting
        header = str(k) + ': {}'.format(size_obj.size) + ' bytes. Detailed size:'
        if len(header) < indent:
            header += ' ' * (indent - len(header))
        file.write(header)

        size_obj = next(r for r in size_obj.refs if r.name == '__dict__')
        # Sort in descending order to select most 'heavy' collections
        for num, sub_obj in enumerate(sorted(size_obj.refs, key=lambda v: v.size, reverse=True)):
            if num > 10:
                break
            file.write('[{} : {}],      '.format(sub_obj.name, sub_obj.size))
        file.write('\n')
    file.close()


@pytest.mark.skip('Unskip if you need to debug')
def test_requests_collection_debugging(looper,
                                       nodeSet,
                                       sdk_wallet_trustee):
    primary = nodeSet[0]

    seed = randomString(32)
    alias = randomString(5)
    wh, _ = sdk_wallet_trustee
    nym_request, new_did = looper.loop.run_until_complete(
        prepare_nym_request(sdk_wallet_trustee, seed,
                            alias, STEWARD_STRING))

    nym_request = json.loads(nym_request)
    a = sys.getsizeof(primary.requests)

    mas = []
    for _ in range(50000):
        req = sdk_json_to_request_object(nym_request)
        req.reqId = randomString(32)
        mas.append(req)
        primary.requests.add_propagate(req, 'asd')
        primary.requests.mark_as_forwarded(req, 2)
        primary.requests.set_finalised(req)

    b = sys.getsizeof(primary.requests)
    lb = len(primary.requests)

    for req in mas:
        primary.requests.mark_as_executed(req)
        primary.requests.free(req.key)
        primary.requests.free(req.key)

    c = sys.getsizeof(primary.requests)
    lc = len(primary.requests)

    for _ in range(100000):
        req = sdk_json_to_request_object(nym_request)
        req.reqId = randomString(32)
        mas.append(req)
        primary.requests.add_propagate(req, 'asd')
        primary.requests.mark_as_forwarded(req, 2)
        primary.requests.set_finalised(req)

    d = sys.getsizeof(primary.requests)
    ld = len(primary.requests)

    print(a)
    print(b, lb)
    print(c, lc)
    print(d, ld)



import pytest

from plenum.common.exceptions import RequestRejectedException, \
    RequestNackedException

from indy_common.constants import POOL_RESTART, ACTION, START, DATETIME
from plenum.common.constants import TXN_TYPE, STEWARD_STRING
from plenum.test.helper import sdk_gen_request, sdk_sign_and_submit_req_obj, \
    sdk_get_reply, sdk_get_and_check_replies


def test_fail_pool_restart_with_steward_role(
        sdk_pool_handle, sdk_wallet_steward, looper):
    op = {
        TXN_TYPE: POOL_RESTART,
        ACTION: START,
    }
    req_obj = sdk_gen_request(op, identifier=sdk_wallet_steward[1])
    req = sdk_sign_and_submit_req_obj(looper,
                                      sdk_pool_handle,
                                      sdk_wallet_steward,
                                      req_obj)
    with pytest.raises(RequestRejectedException) as excinfo:
        sdk_get_and_check_replies(looper, [req], 100)
    assert excinfo.match('Not enough TRUSTEE signatures')


def test_fail_pool_restart_with_invalid_datetime(
        sdk_pool_handle, sdk_wallet_steward, looper):
    invalid_datetime = "12.05.2018 4/40"
    op = {
        TXN_TYPE: POOL_RESTART,
        ACTION: START,
        DATETIME: invalid_datetime
    }
    req_obj = sdk_gen_request(op, identifier=sdk_wallet_steward[1])
    req = sdk_sign_and_submit_req_obj(looper,
                                      sdk_pool_handle,
                                      sdk_wallet_steward,
                                      req_obj)
    with pytest.raises(RequestNackedException) as excinfo:
        sdk_get_and_check_replies(looper, [req], 100)
    assert excinfo.match("datetime " + invalid_datetime + " is not valid")




import pytest

from indy_common.node_version_fallback import (
    InvalidVersionError, NodeVersionFallback,
)


def test_node_version_fallback_str():
    version = '1.2.3'
    assert str(NodeVersionFallback(version)) == version


def test_node_version_fallback_repr():
    version = '1.2.3'
    assert (repr(NodeVersionFallback(version)) == "{}(version='{}')".format(NodeVersionFallback.__name__, version))


@pytest.mark.parametrize(
    'version',
    [
        1, [], {},
        '1',
        '1.2',
        '1.2.3.4',
        '1:1.2.3',
        '1.2.3+4',
        '1.2.3post4',
        '1.2.3dev',
        '1.2.3rc',
        '1.2.3a',
        '1.2.3b',
        # valid PEP440:
        #  alpha prerelease
        #  beta prerelease
        #  postrelease
        #  epoch
        #  local version
        #  parts num != 3
        '1.2.3a1',
        '1.2.3b2',
        '1.2.3.post1',
        '1!1.2.3',
        '1.2.3+1',
        '1',
        '1.2',
        '1.2.3.4'

    ]
)
def test_node_version_fallback_init_invalid(version):
    with pytest.raises(InvalidVersionError):
        NodeVersionFallback(version)


@pytest.mark.parametrize(
    'version',
    [
        '1.2.3',
        '1.2.3.dev1',
        '1.2.3dev1',
        '1.2.3.dev.1',
        '1.2.3.rc1',
        '1.2.3rc1',
        '1.2.3.rc.1',
    ]
)
def test_node_version_fallback_init_valid(version):
    NodeVersionFallback(version)


def test_sem_ver_base_api():
    assert NodeVersionFallback('1.2.3').major == 1
    assert NodeVersionFallback('1.2.3').minor == 2
    assert NodeVersionFallback('1.2.3').patch == 3


def test_node_version_fallback_comparison_operators():
    with pytest.raises(TypeError):
        assert NodeVersionFallback('1.2.2') < NodeVersionFallback('1.2.3')
    with pytest.raises(TypeError):
        assert NodeVersionFallback('1.2.3') > NodeVersionFallback('1.2.2')
    with pytest.raises(TypeError):
        assert NodeVersionFallback('1.2.2') <= NodeVersionFallback('1.2.3')
    with pytest.raises(TypeError):
        assert NodeVersionFallback('1.2.3') <= NodeVersionFallback('1.2.3')
    with pytest.raises(TypeError):
        assert NodeVersionFallback('1.2.3') >= NodeVersionFallback('1.2.2')
    with pytest.raises(TypeError):
        assert NodeVersionFallback('1.2.2') >= NodeVersionFallback('1.2.2')

    assert NodeVersionFallback('1.2.2') == NodeVersionFallback('1.2.2')
    assert NodeVersionFallback('1.2.2') != NodeVersionFallback('1.2.3')


def test_node_version_fallback_upstream():
    pv = NodeVersionFallback('1.2.3')
    assert pv.upstream is pv


def test_node_version_fallback_public():
    assert NodeVersionFallback('1.2.3.dev0').parts == (1, 2, 3, 'dev', 0)
    assert NodeVersionFallback('1.2.3.dev2').parts == (1, 2, 3, 'dev', 2)
    assert NodeVersionFallback('1.2.3.rc0').parts == (1, 2, 3, 'rc', 0)
    assert NodeVersionFallback('1.2.3.rc3').parts == (1, 2, 3, 'rc', 3)
    assert NodeVersionFallback('1.2.3').parts == (1, 2, 3, None, None)


def test_node_version_fallback_parts():
    assert NodeVersionFallback('1.2.3.dev2').parts == (1, 2, 3, 'dev', 2)
    assert NodeVersionFallback('1.2.3.rc3').parts == (1, 2, 3, 'rc', 3)
    assert NodeVersionFallback('1.2.3').parts == (1, 2, 3, None, None)


@pytest.mark.parametrize(
    'version',
    [
        '1.2.3',
        '1.2.3.dev2',
        '1.2.3rc1',
    ]
)
def test_node_version_fallback_base_api(version):
    v = NodeVersionFallback(version)
    assert v.full == version
    assert v.public == version
    assert v.release_parts == v.parts[:3]
    assert v.release == '.'.join(map(str, v.release_parts))



from typing import Optional

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit

from indy_common.config_util import getConfig

from indy_common.constants import CONFIG_LEDGER_ID, POOL_UPGRADE, \
    ACTION, CANCEL, START, SCHEDULE, PACKAGE, REINSTALL

from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_node.server.upgrader import Upgrader
from plenum.common.constants import FORCE, VERSION, NAME
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import get_request_data, get_payload_data
from plenum.server.database_manager import DatabaseManager
from plenum.server.pool_manager import TxnPoolManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler


class PoolUpgradeHandler(WriteRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 upgrader: Upgrader,
                 write_req_validator: WriteRequestValidator,
                 pool_manager: TxnPoolManager):
        super().__init__(database_manager, POOL_UPGRADE, CONFIG_LEDGER_ID)
        self.upgrader = upgrader
        self.write_req_validator = write_req_validator
        self.pool_manager = pool_manager

    def static_validation(self, request: Request):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        action = operation.get(ACTION)
        if action not in (START, CANCEL):
            raise InvalidClientRequest(identifier, req_id,
                                       "{} not a valid action".
                                       format(action))
        if action == START:
            schedule = operation.get(SCHEDULE, {})
            force = operation.get(FORCE)
            force = str(force) == 'True'
            isValid, msg = self.upgrader.isScheduleValid(
                schedule, self.pool_manager.getNodesServices(), force)
            if not isValid:
                raise InvalidClientRequest(identifier, req_id,
                                           "{} not a valid schedule since {}".
                                           format(schedule, msg))

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        status = '*'

        pkg_to_upgrade = operation.get(PACKAGE, getConfig().UPGRADE_ENTRY)
        targetVersion = operation[VERSION]
        reinstall = operation.get(REINSTALL, False)

        if not pkg_to_upgrade:
            raise InvalidClientRequest(identifier, req_id, "Upgrade package name is empty")

        try:
            res = self.upgrader.check_upgrade_possible(pkg_to_upgrade, targetVersion, reinstall)
        except Exception as exc:
            res = str(exc)

        if res:
            raise InvalidClientRequest(identifier, req_id, res)

        action = operation.get(ACTION)
        # TODO: Some validation needed for making sure name and version
        # present
        txn = self.upgrader.get_upgrade_txn(
            lambda txn: get_payload_data(txn).get(
                NAME,
                None) == operation.get(
                NAME,
                None) and get_payload_data(txn).get(VERSION) == operation.get(VERSION),
            reverse=True)
        if txn:
            status = get_payload_data(txn).get(ACTION, '*')

        if status == START and action == START:
            raise InvalidClientRequest(
                identifier,
                req_id,
                "Upgrade '{}' is already scheduled".format(
                    operation.get(NAME)))
        if status == '*':
            auth_action = AuthActionAdd(txn_type=POOL_UPGRADE,
                                        field=ACTION,
                                        value=action)
        else:
            auth_action = AuthActionEdit(txn_type=POOL_UPGRADE,
                                         field=ACTION,
                                         old_value=status,
                                         new_value=action)
        self.write_req_validator.validate(request,
                                          [auth_action])

    def apply_forced_request(self, req: Request):
        super().apply_forced_request(req)
        txn = self._req_to_txn(req)
        self.upgrader.handleUpgradeTxn(txn)

    # Config handler don't use state for any validation for now
    def update_state(self, txn, prev_result, request, is_committed=False):
        pass



from copy import deepcopy

from indy_common.constants import SHA256
from indy_node.test import waits
from indy_node.test.upgrade.helper import bumpVersion, checkUpgradeScheduled, \
    sdk_ensure_upgrade_sent
from plenum.common.constants import VERSION, NAME
from plenum.common.util import randomString
from stp_core.loop.eventually import eventually


def testRescheduleUpgradeToLowerVersionThanPreviouslyScheduled(
        looper, tconf, nodeSet, validUpgrade, sdk_pool_handle, sdk_wallet_trustee):
    """
    A node starts at version 1.2 running has scheduled upgrade for version 1.5
    but get a txn for upgrade 1.4, it will schedule it and cancel upgrade to 1.5.
    """
    upgr1 = deepcopy(validUpgrade)

    upgr2 = deepcopy(upgr1)
    upgr2[VERSION] = bumpVersion(upgr1[VERSION])
    upgr2[NAME] += randomString(3)
    # upgr2[SHA256] = get_valid_code_hash()
    upgr2[SHA256] = 'ef9c3984e7a31994d4f692139116120bd0dd1ff7e270b6a2d773f8f2f9214d4c'

    # An upgrade for higher version scheduled, it should pass
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgr2)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr2[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))

    # An upgrade for lower version scheduled, the transaction should pass and
    # the upgrade should be scheduled
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgr1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr1[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))



import pytest
import time
import json

from indy.anoncreds import issuer_create_and_store_credential_def
from indy.ledger import build_cred_def_request

from indy_common.state.domain import make_state_path_for_claim_def
from indy_node.test.anon_creds.helper import get_cred_def_id, create_revoc_reg, create_revoc_reg_entry
from plenum.common.util import randomString
from indy_common.constants import REVOC_REG_ENTRY, REVOC_REG_DEF_ID, ISSUED, \
    REVOKED, PREV_ACCUM, ACCUM, REVOC_REG_DEF, ISSUANCE_BY_DEFAULT, \
    CRED_DEF_ID, VALUE, TAG, ISSUANCE_ON_DEMAND, ID, \
    TXN_TYPE, REVOC_TYPE, ISSUANCE_TYPE, MAX_CRED_NUM, TAILS_HASH, TAILS_LOCATION, PUBLIC_KEYS, \
    GET_REVOC_REG, TIMESTAMP, CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_SIGNATURE_TYPE, \
    CLAIM_DEF_TAG
from indy_common.types import Request
from indy_common.state import domain
from plenum.test.helper import sdk_sign_request_from_dict, sdk_send_and_check, sdk_sign_and_submit_req, \
    sdk_get_and_check_replies
from plenum.common.txn_util import reqToTxn, append_txn_metadata
from plenum.common.types import f, OPERATION

from plenum.test.conftest import sdk_wallet_client
from indy_node.test.schema.test_send_get_schema import send_schema_req

CRED_DEF_VERSION = '1.0'
SCHEMA_VERSION = '1.0'


@pytest.fixture(scope="module")
def claim_def(looper, sdk_wallet_handle, sdk_wallet_steward, send_schema_req):
    schema = json.loads(send_schema_req[0])
    tag = 'some_tag'
    schema_seq_no = send_schema_req[1]['result']['txnMetadata']['seqNo']
    schema['seqNo'] = schema_seq_no

    definition_id, definition_json = \
        looper.loop.run_until_complete(issuer_create_and_store_credential_def(
            sdk_wallet_handle, sdk_wallet_steward[1], json.dumps(schema),
            tag, "CL", json.dumps({"support_revocation": True})))
    cred_def = looper.loop.run_until_complete(build_cred_def_request(sdk_wallet_steward[1], definition_json))
    return json.loads(cred_def)['operation']


@pytest.fixture(scope="module")
def send_claim_def(looper,
                   txnPoolNodeSet,
                   sdk_wallet_steward,
                   sdk_pool_handle,
                   claim_def):
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, claim_def)
    rep = sdk_send_and_check([json.dumps(req)], looper, txnPoolNodeSet, sdk_pool_handle)[0]
    return rep


@pytest.fixture(scope="module")
def send_revoc_reg_def(looper, sdk_wallet_handle, sdk_pool_handle, sdk_wallet_steward, send_claim_def):
    revoc_reg = create_revoc_reg(looper, sdk_wallet_handle, sdk_wallet_steward[1], randomString(5),
                                 get_cred_def_id(send_claim_def[0][f.IDENTIFIER.nm],
                                                 send_claim_def[0]['operation']['ref'],
                                                 send_claim_def[0]['operation']['tag']),
                                 ISSUANCE_BY_DEFAULT)
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, revoc_reg)
    rep = sdk_get_and_check_replies(looper, [req])[0]
    return rep


@pytest.fixture(scope="module")
def send_revoc_reg_entry(looper, sdk_wallet_handle, sdk_pool_handle, sdk_wallet_steward, send_claim_def):
    revoc_reg, revoc_entry = create_revoc_reg_entry(looper, sdk_wallet_handle, sdk_wallet_steward[1], randomString(5),
                                                    get_cred_def_id(send_claim_def[0][f.IDENTIFIER.nm],
                                                                    send_claim_def[0]['operation']['ref'],
                                                                    send_claim_def[0]['operation']['tag']),
                                                    ISSUANCE_BY_DEFAULT)
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, revoc_reg)
    revoc_reg = sdk_get_and_check_replies(looper, [req])[0]
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, revoc_entry)
    revoc_entry = sdk_get_and_check_replies(looper, [req])[0]
    return revoc_reg, revoc_entry


@pytest.fixture(scope="module")
def add_revoc_def_by_default(create_node_and_not_start,
                             looper,
                             sdk_wallet_steward):
    node = create_node_and_not_start
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)

    txn = append_txn_metadata(reqToTxn(Request(**req)),
                              txn_time=int(time.time()),
                              seq_no=node.domainLedger.seqNo + 1)
    node.write_manager.update_state(txn)
    return req


def build_revoc_reg_entry_for_given_revoc_reg_def(
        revoc_def_req):
    path = ":".join([revoc_def_req[f.IDENTIFIER.nm],
                     domain.MARKER_REVOC_DEF,
                     revoc_def_req[OPERATION][CRED_DEF_ID],
                     revoc_def_req[OPERATION][REVOC_TYPE],
                     revoc_def_req[OPERATION][TAG]])
    data = {
        REVOC_REG_DEF_ID: path,
        TXN_TYPE: REVOC_REG_ENTRY,
        REVOC_TYPE: revoc_def_req[OPERATION][REVOC_TYPE],
        VALUE: {
            PREV_ACCUM: randomString(10),
            ACCUM: randomString(10),
            ISSUED: [],
            REVOKED: [],
        }
    }
    return data


@pytest.fixture(scope="function")
def build_txn_for_revoc_def_entry_by_default(looper,
                                             sdk_wallet_steward,
                                             add_revoc_def_by_default):
    revoc_def_req = add_revoc_def_by_default
    data = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req)
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return req


@pytest.fixture(scope="module")
def add_revoc_def_by_demand(create_node_and_not_start,
                            looper,
                            sdk_wallet_steward):
    node = create_node_and_not_start
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_ON_DEMAND,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)

    txn = append_txn_metadata(reqToTxn(Request(**req)),
                              txn_time=int(time.time()),
                              seq_no=node.domainLedger.seqNo + 1)
    node.write_manager.update_state(txn)
    return req


@pytest.fixture(scope="module")
def build_txn_for_revoc_def_entry_by_demand(looper,
                                            sdk_wallet_steward,
                                            add_revoc_def_by_demand):
    revoc_def_req = add_revoc_def_by_demand
    path = ":".join([revoc_def_req[f.IDENTIFIER.nm],
                     domain.MARKER_REVOC_DEF,
                     revoc_def_req[OPERATION][CRED_DEF_ID],
                     revoc_def_req[OPERATION][REVOC_TYPE],
                     revoc_def_req[OPERATION][TAG]])
    data = {
        REVOC_REG_DEF_ID: path,
        TXN_TYPE: REVOC_REG_ENTRY,
        VALUE: {
            PREV_ACCUM: randomString(10),
            ACCUM: randomString(10),
            ISSUED: [],
            REVOKED: [],
        }
    }

    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return req


@pytest.fixture(scope="module")
def build_revoc_def_by_default(looper, sdk_wallet_steward):
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return req


@pytest.fixture(scope="module")
def build_revoc_def_by_endorser(looper, sdk_wallet_endorser):
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_endorser, data)
    return req


@pytest.fixture(scope="module")
def build_revoc_def_by_steward(looper, sdk_wallet_steward):
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return req


def build_revoc_def_random(looper, sdk_wallet):
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_BY_DEFAULT,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet, data)
    return req


@pytest.fixture(scope="module")
def build_revoc_def_by_demand(looper, sdk_wallet_steward):
    data = {
        ID: randomString(50),
        TXN_TYPE: REVOC_REG_DEF,
        REVOC_TYPE: "CL_ACCUM",
        TAG: randomString(5),
        CRED_DEF_ID: ":".join(4 * [randomString(10)]),
        VALUE: {
            ISSUANCE_TYPE: ISSUANCE_ON_DEMAND,
            MAX_CRED_NUM: 1000000,
            TAILS_HASH: randomString(50),
            TAILS_LOCATION: 'http://tails.location.com',
            PUBLIC_KEYS: {},
        }
    }
    req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, data)
    return req


@pytest.fixture(scope="module")
def send_revoc_reg_def_by_default(looper,
                                  txnPoolNodeSet,
                                  sdk_wallet_steward,
                                  sdk_pool_handle,
                                  send_claim_def,
                                  build_revoc_def_by_default):
    _, author_did = sdk_wallet_steward
    claim_def_req = send_claim_def[0]
    revoc_reg = build_revoc_def_by_default
    revoc_reg['operation'][CRED_DEF_ID] = \
        make_state_path_for_claim_def(author_did,
                                      str(claim_def_req['operation'][CLAIM_DEF_SCHEMA_REF]),
                                      claim_def_req['operation'][CLAIM_DEF_SIGNATURE_TYPE],
                                      claim_def_req['operation'][CLAIM_DEF_TAG]
                                      ).decode()
    revoc_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, revoc_reg['operation'])
    _, revoc_reply = sdk_send_and_check([json.dumps(revoc_req)], looper, txnPoolNodeSet, sdk_pool_handle)[0]
    return revoc_req, revoc_reply


@pytest.fixture(scope="module")
def send_revoc_reg_def_by_demand(looper,
                                 txnPoolNodeSet,
                                 sdk_wallet_steward,
                                 sdk_pool_handle,
                                 send_claim_def,
                                 build_revoc_def_by_demand):
    _, author_did = sdk_wallet_steward
    claim_def_req = send_claim_def[0]
    revoc_reg = build_revoc_def_by_demand
    revoc_reg['operation'][CRED_DEF_ID] = make_state_path_for_claim_def(author_did,
                                                                        str(claim_def_req['operation'][
                                                                                CLAIM_DEF_SCHEMA_REF]),
                                                                        claim_def_req['operation'][
                                                                            CLAIM_DEF_SIGNATURE_TYPE],
                                                                        claim_def_req['operation'][CLAIM_DEF_TAG]
                                                                        ).decode()
    revoc_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, revoc_reg['operation'])
    sdk_send_and_check([json.dumps(revoc_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    return revoc_req


@pytest.fixture(scope="module")
def send_revoc_reg_entry_by_default(looper,
                                    txnPoolNodeSet,
                                    sdk_wallet_steward,
                                    sdk_pool_handle,
                                    send_revoc_reg_def_by_default):
    revoc_def_req, _ = send_revoc_reg_def_by_default
    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req)
    rev_reg_entry[VALUE][REVOKED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, rev_reg_entry)
    reg_entry_replies = sdk_send_and_check([json.dumps(rev_entry_req)],
                                           looper,
                                           txnPoolNodeSet,
                                           sdk_pool_handle)
    return reg_entry_replies[0]


@pytest.fixture(scope="module")
def send_revoc_reg_entry_by_demand(looper,
                                   txnPoolNodeSet,
                                   sdk_wallet_steward,
                                   sdk_pool_handle,
                                   send_revoc_reg_def_by_demand):
    revoc_def_req = send_revoc_reg_def_by_demand
    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req)
    rev_reg_entry[VALUE][ISSUED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, rev_reg_entry)
    reg_entry_replies = sdk_send_and_check([json.dumps(rev_entry_req)],
                                           looper,
                                           txnPoolNodeSet,
                                           sdk_pool_handle)
    return reg_entry_replies[0]


@pytest.fixture(scope="module")
def create_node_and_not_start(create_node_and_not_start):
    create_node_and_not_start.bootstrapper.upload_states()
    return create_node_and_not_start


from typing import Iterable
import re


class InvalidVersionError(ValueError):
    pass


class NodeVersionFallback:
    """ Provide basic PEP440 based validation in raw python environment. """

    # covers only some cases
    re_pep440 = re.compile(r'([0-9]+)\.([0-9]+)\.([0-9]+)(?:\.?(dev|rc)\.?([0-9]+))?')

    def __init__(self, version: str):
        try:
            match = self.re_pep440.fullmatch(version)
        except TypeError as exc:
            raise InvalidVersionError(str(version)) from exc

        if not match:
            raise InvalidVersionError(
                "version '{}' is invalid, expected N.N.N[[.]{{dev|rc}}[.]N]"
                .format(version)
            )
        self._version = tuple(
            [int(p) if idx in (0, 1, 2, 4) and p is not None else p for idx, p in enumerate(match.groups())]
        )

    @property
    def public(self):
        return "{}.{}.{}".format(*self._version[:3]) + (
            '' if self._version[3] is None else
            "{}{}{}"
            .format(
                '.' if self._version[3] == 'dev' else '', *self._version[3:]
            )
        )

    @property
    def full(self) -> str:
        return self.public

    @property
    def parts(self) -> Iterable:
        return self._version

    @property
    def release(self) -> str:
        return '.'.join(map(str, self.release_parts))

    @property
    def release_parts(self) -> Iterable:
        return self._version[:3]

    @property
    def major(self) -> str:
        return self.release_parts[0]

    @property
    def minor(self) -> str:
        return self.release_parts[1]

    @property
    def patch(self) -> str:
        return self.release_parts[2]

    @property
    def upstream(self) -> 'NodeVersionFallback':
        """ Upstream part of the package. """
        return self

    def __str__(self):
        return self.full

    def __repr__(self):
        return "{}(version='{}')".format(self.__class__.__name__, self.full)

    def __hash__(self):
        return hash(self.full)

    def __eq__(self, other: 'NodeVersionFallback') -> bool:
        return self._version == other._version

    def __ne__(self, other: 'NodeVersionFallback') -> bool:
        return not (self == other)



from indy_common.constants import GET_RICH_SCHEMA_OBJECT_BY_ID, RS_ID
from indy_common.config_util import getConfig
from indy_node.server.request_handlers.read_req_handlers.rich_schema.abstract_rich_schema_read_req_handler import \
    AbstractRichSchemaReadRequestHandler
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.common.exceptions import InvalidClientRequest
from plenum.server.database_manager import DatabaseManager


class GetRichSchemaObjectByIdHandler(AbstractRichSchemaReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, GET_RICH_SCHEMA_OBJECT_BY_ID, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        super().get_result(request)
        self._validate_request_type(request)

        id = request.operation[RS_ID]

        try:
            value, seq_no, last_update_time, proof = self.lookup(id, is_committed=True, with_proof=True)
        except KeyError:
            value, seq_no, last_update_time, proof = None, None, None, None

        return self.make_result(request=request,
                                data=value,
                                last_seq_no=seq_no,
                                update_time=last_update_time,
                                proof=proof)



from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_node.server.request_handlers.action_req_handlers.utils import generate_action_result
from indy_node.server.restarter import Restarter

from indy_common.constants import POOL_RESTART, ACTION

from plenum.common.request import Request
from plenum.common.txn_util import get_request_data
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.action_request_handler import ActionRequestHandler
from stp_core.common.log import getlogger

logger = getlogger()


class PoolRestartHandler(ActionRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator,
                 restarter: Restarter):
        super().__init__(database_manager, POOL_RESTART, None)
        self.write_req_validator = write_req_validator
        self.restarter = restarter

    def static_validation(self, request: Request):
        self._validate_request_type(request)

    def dynamic_validation(self, request: Request):
        self._validate_request_type(request)
        action = request.operation.get(ACTION)
        self.write_req_validator.validate(request,
                                          [AuthActionAdd(txn_type=POOL_RESTART,
                                                         field=ACTION,
                                                         value=action)])

    def process_action(self, request: Request):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        logger.debug("Transaction {} with type {} started"
                     .format(req_id, request.txn_type))
        self.restarter.handleRestartRequest(request)
        result = generate_action_result(request)
        logger.debug("Transaction {} with type {} finished"
                     .format(req_id, request.txn_type))
        return result



# Some unit tests for upgrader
import pytest

import indy_common
from indy_common.constants import APP_NAME
from indy_common.version import src_version_cls
from indy_node.server.upgrader import Upgrader
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion


some_pkg_name = 'some_package'


class PkgDebianVersion(DebianVersion):
    def __init__(self, pkg_name, version, *args, **kwargs):
        super().__init__(
            version,
            upstream_cls=indy_common.version.src_version_cls(pkg_name),
            *args, **kwargs
        )


@pytest.fixture(autouse=True)
def node_control_util_patched(monkeypatch, request):
    def pkg_version(pkg_name, version):
        return (
            PkgDebianVersion(
                pkg_name or some_pkg_name, version) if version else None
        )

    def pkg_info(pkg_name, version, deps):
        def wrapped(*x):
            return (pkg_version(pkg_name, version), deps)
        return wrapped

    def latest_pkg_ver(pkg_name, version):
        def wrapped(*x, **y):
            return pkg_version(pkg_name, version)
        return wrapped

    marker = request.node.get_closest_marker('pkg_info')
    if marker:
        assert len(marker.args) > 1
        monkeypatch.setattr(
            NodeControlUtil, 'curr_pkg_info',
            pkg_info(
                marker.kwargs.get('pkg_name'),
                marker.args[0],
                marker.args[1]
            )
        )

    marker = request.node.get_closest_marker('latest_pkg_ver')
    if marker:
        assert len(marker.args) > 0
        monkeypatch.setattr(
            NodeControlUtil, 'get_latest_pkg_version',
            latest_pkg_ver(
                marker.kwargs.get('pkg_name'),
                marker.args[0]
            )
        )


@pytest.mark.parametrize(
    'lower_version,higher_version',
    [
        ('0.0.5', '0.0.6'),
        ('0.1.2', '0.2.6'),
        ('1.10.2', '2.0.6'),
        ('1.2.3.dev1', '1.2.3rc1'),
        ('1.2.3.dev1', '1.2.3'),
        ('1.2.3.rc2', '1.2.3'),
    ]
)
def test_versions_comparison(lower_version, higher_version):
    assert Upgrader.compareVersions(higher_version, lower_version) == 1
    assert Upgrader.compareVersions(lower_version, higher_version) == -1
    assert Upgrader.compareVersions(higher_version, higher_version) == 0

    version_cls = src_version_cls(APP_NAME)
    lower_version = version_cls(lower_version)
    higher_version = version_cls(higher_version)
    assert not Upgrader.is_version_upgradable(higher_version, higher_version)
    assert Upgrader.is_version_upgradable(
        higher_version, higher_version, reinstall=True)
    assert Upgrader.is_version_upgradable(lower_version, higher_version)
    assert not Upgrader.is_version_upgradable(higher_version, lower_version)


def test_get_src_version_for_app(monkeypatch):
    called = 0

    def _f(*args, **kwargs):
        nonlocal called
        called += 1
        return (None, [])

    monkeypatch.setattr(NodeControlUtil, 'curr_pkg_info', _f)
    assert Upgrader.get_src_version(APP_NAME)
    assert not called
    Upgrader.get_src_version(APP_NAME, nocache=True)
    assert called


@pytest.mark.pkg_info('1:1.2.2-3', [])
def test_get_src_version_for(monkeypatch):
    assert (Upgrader.get_src_version(some_pkg_name) ==
            src_version_cls(some_pkg_name)('1.2.2'))


def test_check_upgrade_possible_invalid_target_version():
    assert 'invalid target version' in Upgrader.check_upgrade_possible(
        APP_NAME, '1.2.3.4')


@pytest.mark.pkg_info(None, [])
def test_check_upgrade_possible_pkg_not_installed():
    assert ('is not installed' in Upgrader.check_upgrade_possible(
        some_pkg_name, '1.2.3'))


@pytest.mark.pkg_info('1.2.2', ['pkg1'])
def test_check_upgrade_possible_invalid_top_level_pkg():
    assert ("doesn't belong to pool" in Upgrader.check_upgrade_possible(
        some_pkg_name, '1.2.3'))


@pytest.mark.pkg_info('1.2.2', [APP_NAME])
def test_check_upgrade_possible_not_gt_version():
    assert ("is not upgradable" in Upgrader.check_upgrade_possible(
        some_pkg_name, '1.2.2'))


@pytest.mark.pkg_info('1.2.2', [APP_NAME])
def test_check_upgrade_possible_not_ge_version_reinstall():
    assert ("is not upgradable" in Upgrader.check_upgrade_possible(
        some_pkg_name, '1.2.1', reinstall=True))


@pytest.mark.skip(reason='INDY-2026')
@pytest.mark.pkg_info('1.2.2', [APP_NAME])
#@pytest.mark.latest_pkg_ver(None) TODO INDY-2026
def test_check_upgrade_possible_no_pkg_with_target_version():
    target_ver = '1.2.3'
    assert (
        "for target version {} is not found".format(target_ver) in
        Upgrader.check_upgrade_possible(some_pkg_name, target_ver)
    )


@pytest.mark.pkg_info('1.2.2', [APP_NAME])
#@pytest.mark.latest_pkg_ver('1.2.3') TODO INDY-2026
def test_check_upgrade_possible_succeeded():
    target_ver = '1.2.3'
    assert not Upgrader.check_upgrade_possible(some_pkg_name, target_ver)
    assert not Upgrader.check_upgrade_possible(
        some_pkg_name, target_ver, reinstall=True)


@pytest.mark.pkg_info('1.2.2', [], pkg_name=APP_NAME)
#@pytest.mark.latest_pkg_ver('1.2.3') TODO INDY-2026
def test_check_upgrade_possible_succeeded_for_app_pkg():
    target_ver = '1.2.3'
    assert not Upgrader.check_upgrade_possible(APP_NAME, target_ver)
    assert not Upgrader.check_upgrade_possible(
        APP_NAME, target_ver, reinstall=True)


@pytest.mark.pkg_info('1.2.2', [APP_NAME])
#@pytest.mark.latest_pkg_ver('1.2.2') TODO INDY-2026
def test_check_upgrade_possible_reinstall_succeeded():
    assert not Upgrader.check_upgrade_possible(
        some_pkg_name, '1.2.2', reinstall=True)


@pytest.mark.pkg_info('1.2.2', [], pkg_name=APP_NAME)
#@pytest.mark.latest_pkg_ver('1.2.2') TODO INDY-2026
def test_check_upgrade_possible_reinstall_succeeded_for_app_pkg():
    assert not Upgrader.check_upgrade_possible(
        APP_NAME, '1.2.2', reinstall=True)



from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER, AuthConstraintOr, AuthConstraintAnd, \
    AuthConstraintForbidden
from indy_common.constants import ENDORSER
from indy_common.test.auth.metadata.helper import validate, PLUGIN_FIELD, Action
from plenum.common.constants import TRUSTEE, STEWARD

MAX_SIG_COUNT = 3


def test_plugin_and_or_rule_same_role_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                          signatures, amount):
    validate(
        auth_constraint=AuthConstraintAnd(auth_constraints=[
            AuthConstraintOr(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=2, need_to_be_owner=False),
                AuthConstraint(role=TRUSTEE, sig_count=2, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 2}),
            ]),
            AuthConstraintOr(auth_constraints=[
                AuthConstraint(role=STEWARD, sig_count=3, need_to_be_owner=False),
                AuthConstraint(role=STEWARD, sig_count=3, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 2}),
            ])
        ]),
        valid_actions=[
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 2, STEWARD: 3},
                   is_owner=False, amount=2, extra_sigs=True),
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, STEWARD: 3},
                   is_owner=False, amount=2, extra_sigs=True),

            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 2, STEWARD: 3},
                   is_owner=False, amount=None, extra_sigs=True),
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, STEWARD: 3},
                   is_owner=False, amount=None, extra_sigs=True),
        ],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=False, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_and_or_rule_diff_role_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                          signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraintAnd(auth_constraints=[
            AuthConstraintOr(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=2, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 3}),
                AuthConstraint(role=STEWARD, sig_count=2, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 3}),
            ]),
            AuthConstraintOr(auth_constraints=[
                AuthConstraint(role=ENDORSER, sig_count=3, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 3}),
                AuthConstraint(role=IDENTITY_OWNER, sig_count=3, need_to_be_owner=True,
                               metadata={PLUGIN_FIELD: 3}),
            ])
        ]),
        valid_actions=[Action(author=TRUSTEE, endorser=None, sigs={ENDORSER: 3, TRUSTEE: s},
                              is_owner=owner, amount=3, extra_sigs=True)
                       for s in range(2, 4)
                       for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={STEWARD: s1,
                                                        ENDORSER: 3,
                                                        TRUSTEE: s2},
                   is_owner=owner, amount=3, extra_sigs=True)
            for s1 in range(2, 4)
            for s2 in range(1, MAX_SIG_COUNT + 1)
            for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={STEWARD: s1,
                                                        IDENTITY_OWNER: 3,
                                                        TRUSTEE: s2},
                   is_owner=True, amount=3, extra_sigs=True)
            for s1 in range(2, 4)
            for s2 in range(1, MAX_SIG_COUNT + 1)],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_or_and_rule_diff_roles_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                           signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraintOr(auth_constraints=[
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False),
            ]),
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=1, need_to_be_owner=False),
                AuthConstraint(role=STEWARD, sig_count=2, need_to_be_owner=False),
            ]),
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=1, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 2}),
                AuthConstraint(role=ENDORSER, sig_count=2, need_to_be_owner=True,
                               metadata={PLUGIN_FIELD: 2}),
            ]),
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=2, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 3}),
                AuthConstraint(role=IDENTITY_OWNER, sig_count=3, need_to_be_owner=True,
                               metadata={PLUGIN_FIELD: 3}),
            ]),
        ]),
        valid_actions=[Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3},
                              is_owner=False, amount=None, extra_sigs=True),
                       Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3},
                              is_owner=True, amount=None, extra_sigs=True)] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: s1, STEWARD: s2},
                   is_owner=owner, amount=None, extra_sigs=True)
            for s1 in range(1, 4)
            for s2 in range(2, 4)
            for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: s1, ENDORSER: s2},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, 4) for s2 in range(2, 4)] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 2, IDENTITY_OWNER: 3},
                   is_owner=True, amount=3, extra_sigs=True),
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, IDENTITY_OWNER: 3},
                   is_owner=True, amount=3, extra_sigs=True)],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_complex_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                            signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraintOr(auth_constraints=[
            # 1st
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 1}),
                AuthConstraintOr(auth_constraints=[
                    AuthConstraintAnd(auth_constraints=[
                        AuthConstraint(role=STEWARD, sig_count=1, need_to_be_owner=False,
                                       metadata={PLUGIN_FIELD: 1}),
                        AuthConstraint(role=ENDORSER, sig_count=1, need_to_be_owner=False,
                                       metadata={PLUGIN_FIELD: 1}),
                    ]),
                    AuthConstraint(role=STEWARD, sig_count=2, need_to_be_owner=False,
                                   metadata={PLUGIN_FIELD: 1}),
                    AuthConstraint(role=ENDORSER, sig_count=2, need_to_be_owner=False,
                                   metadata={PLUGIN_FIELD: 1}),
                ])
            ]),
            # 2d
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 3}),
                AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True,
                               metadata={PLUGIN_FIELD: 3})
            ]),
            # 3d
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False),
                AuthConstraintOr(auth_constraints=[
                    AuthConstraint(role=STEWARD, sig_count=2, need_to_be_owner=False),
                    AuthConstraint(role=ENDORSER, sig_count=2, need_to_be_owner=True),
                ])
            ]),
            # 4th
            AuthConstraintAnd(auth_constraints=[
                AuthConstraint(role=TRUSTEE, sig_count=2, need_to_be_owner=False,
                               metadata={PLUGIN_FIELD: 2}),
                AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True,
                               metadata={PLUGIN_FIELD: 2})
            ]),
        ]),
        valid_actions=[Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, STEWARD: s1, ENDORSER: s2},  # 1st
                              is_owner=owner, amount=1, extra_sigs=True)
                       for s1 in range(1, 4)
                       for s2 in range(1, 4)
                       for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, STEWARD: s},
                   is_owner=owner, amount=1, extra_sigs=True)
            for s in range(2, 4)
            for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, IDENTITY_OWNER: s},  # 2d
                   is_owner=True, amount=3, extra_sigs=True)
            for s in range(1, 4)] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, STEWARD: s},  # 3d
                   is_owner=owner, amount=None, extra_sigs=True)
            for s in range(2, 4)
            for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3, ENDORSER: s},
                   is_owner=owner, amount=None, extra_sigs=True)
            for s in range(2, 4)
            for owner in [True, False]] + [
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 2, IDENTITY_OWNER: s},  # 4th
                   is_owner=True, amount=2, extra_sigs=True)
            for s in range(1, 4)],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_complex_with_and_rule_with_not_allowed(write_auth_req_validator, write_request_validation,
                                                       signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraintAnd(auth_constraints=[
            AuthConstraintForbidden(),
            AuthConstraint(role="*", sig_count=1, need_to_be_owner=False,
                           off_ledger_signature=off_ledger_signature,
                           metadata={PLUGIN_FIELD: 2})
        ]),
        valid_actions=[],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_complex_with_or_rule_with_not_allowed_trustee_no_endorser(write_auth_req_validator,
                                                                          write_request_validation,
                                                                          signatures, is_owner, off_ledger_signature,
                                                                          amount):
    validate(
        auth_constraint=AuthConstraintOr(auth_constraints=[
            AuthConstraintForbidden(),
            AuthConstraint(role="*", sig_count=1, need_to_be_owner=False,
                           off_ledger_signature=off_ledger_signature,
                           metadata={PLUGIN_FIELD: 2})
        ]),
        valid_actions=[
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: s1},
                   is_owner=owner, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for owner in [True, False]
        ],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )



import os
import shutil

from indy_common.config_helper import NodeConfigHelper
from indy_common.config_util import getConfig
from stp_core.common.log import getlogger


logger = getlogger()

ENV_FILE_PATH = "/etc/indy/indy.env"
BUILDER_NET_NETWORK_NAME = "net3"


def get_node_name():
    node_name = None
    node_name_key = 'NODE_NAME'

    if os.path.exists(ENV_FILE_PATH):
        with open(ENV_FILE_PATH, "r") as fenv:
            for line in fenv.readlines():
                if line.find(node_name_key) != -1:
                    node_name = line.split('=')[1].strip()
                    break
    if node_name is None:
        logger.error("{} file doesn't contains a node name. Please add string like NODE_NAME=<node name> into this file".format(ENV_FILE_PATH))

    return node_name


def remove_dir(path_to_dir):
    try:
        shutil.rmtree(path_to_dir)
    except shutil.Error as ex:
        logger.error("""While removing directory: {} the next error was raised:
{}""".format(path_to_dir, ex))

        return False
    return True


def migrate_all():

    node_name = get_node_name()
    if node_name is None:
        return False

    config = getConfig()
    config_helper = NodeConfigHelper(node_name, config)

    if BUILDER_NET_NETWORK_NAME != config.NETWORK_NAME:
        logger.info("This script can be used only for {} network".format(BUILDER_NET_NETWORK_NAME))
        return False

    path_to_config_state = os.path.join(config_helper.ledger_dir, config.configStateDbName)
    path_to_config_ts_db = os.path.join(config_helper.ledger_dir, config.configStateTsDbName)

    if not os.path.exists(path_to_config_ts_db):
        logger.error("Path {} to config's timestamp storage does not exist".format(path_to_config_ts_db))
        return False

    if not os.path.exists(path_to_config_state):
        logger.error("Path {} to config_state storage does not exist".format(path_to_config_state))
        return False

    if not remove_dir(path_to_config_ts_db):
        logger.error("Failed to remove {}".format(path_to_config_ts_db))
        return False

    if not remove_dir(path_to_config_state):
        logger.error("Failed to remove {}".format(path_to_config_state))
        return False

    logger.info("Config state storage was successfully removed. Path was {}".format(path_to_config_state))

    return True


if not migrate_all():
    logger.error("Config state storage removing failed")



from collections import OrderedDict
from typing import Dict

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_constraints import AuthConstraint, AuthConstraintOr, accepted_roles, IDENTITY_OWNER, \
    AuthConstraintForbidden
from indy_common.constants import ENDORSER, POOL_CONFIG, VALIDATOR_INFO, POOL_UPGRADE, POOL_RESTART, NODE, \
    CLAIM_DEF, SCHEMA, NYM, ROLE, AUTH_RULE, NETWORK_MONITOR, REVOC_REG_ENTRY, \
    REVOC_REG_DEF, ATTRIB, AUTH_RULES, JSON_LD_CONTEXT, RICH_SCHEMA, RICH_SCHEMA_MAPPING, \
    RICH_SCHEMA_ENCODING, RICH_SCHEMA_CRED_DEF, RICH_SCHEMA_PRES_DEF
from plenum.common.constants import TRUSTEE, STEWARD, VERKEY, TXN_AUTHOR_AGREEMENT, TXN_AUTHOR_AGREEMENT_AML, \
    TXN_AUTHOR_AGREEMENT_DISABLE, LEDGERS_FREEZE

edit_role_actions = {}  # type: Dict[str, Dict[str, AuthActionEdit]]
for role_from in accepted_roles:
    edit_role_actions[role_from] = {}
    for role_to in accepted_roles:
        edit_role_actions[role_from][role_to] = AuthActionEdit(txn_type=NYM,
                                                               field=ROLE,
                                                               old_value=role_from,
                                                               new_value=role_to)

add_new_trustee = AuthActionAdd(txn_type=NYM,
                                field=ROLE,
                                value=TRUSTEE)

add_new_steward = AuthActionAdd(txn_type=NYM,
                                field=ROLE,
                                value=STEWARD)

add_new_endorser = AuthActionAdd(txn_type=NYM,
                                 field=ROLE,
                                 value=ENDORSER)

add_new_network_monitor = AuthActionAdd(txn_type=NYM,
                                        field=ROLE,
                                        value=NETWORK_MONITOR)

add_new_identity_owner = AuthActionAdd(txn_type=NYM,
                                       field=ROLE,
                                       value=IDENTITY_OWNER)

key_rotation = AuthActionEdit(txn_type=NYM,
                              field=VERKEY,
                              old_value='*',
                              new_value='*')

txn_author_agreement = AuthActionAdd(txn_type=TXN_AUTHOR_AGREEMENT,
                                     field='*',
                                     value='*')

edit_txn_author_agreement = AuthActionEdit(txn_type=TXN_AUTHOR_AGREEMENT,
                                           field='*',
                                           old_value='*',
                                           new_value='*')

disable_txn_author_agreement = AuthActionAdd(txn_type=TXN_AUTHOR_AGREEMENT_DISABLE,
                                             field='*',
                                             value='*')

txn_author_agreement_aml = AuthActionAdd(txn_type=TXN_AUTHOR_AGREEMENT_AML,
                                         field='*',
                                         value='*')

add_attrib = AuthActionAdd(txn_type=ATTRIB,
                           field='*',
                           value='*')

edit_attrib = AuthActionEdit(txn_type=ATTRIB,
                             field='*',
                             old_value='*',
                             new_value='*')

add_schema = AuthActionAdd(txn_type=SCHEMA,
                           field='*',
                           value='*')

edit_schema = AuthActionEdit(txn_type=SCHEMA,
                             field='*',
                             old_value='*',
                             new_value='*')

add_json_ld_context = AuthActionAdd(txn_type=JSON_LD_CONTEXT,
                                    field='*',
                                    value='*')

edit_json_ld_context = AuthActionEdit(txn_type=JSON_LD_CONTEXT,
                                      field='*',
                                      old_value='*',
                                      new_value='*')

add_rich_schema = AuthActionAdd(txn_type=RICH_SCHEMA,
                                field='*',
                                value='*')

edit_rich_schema = AuthActionEdit(txn_type=RICH_SCHEMA,
                                  field='*',
                                  old_value='*',
                                  new_value='*')

add_rich_schema_encoding = AuthActionAdd(txn_type=RICH_SCHEMA_ENCODING,
                                         field='*',
                                         value='*')

edit_rich_schema_encoding = AuthActionEdit(txn_type=RICH_SCHEMA_ENCODING,
                                           field='*',
                                           old_value='*',
                                           new_value='*')

add_rich_schema_mapping = AuthActionAdd(txn_type=RICH_SCHEMA_MAPPING,
                                        field='*',
                                        value='*')

edit_rich_schema_mapping = AuthActionEdit(txn_type=RICH_SCHEMA_MAPPING,
                                          field='*',
                                          old_value='*',
                                          new_value='*')

add_rich_schema_cred_def = AuthActionAdd(txn_type=RICH_SCHEMA_CRED_DEF,
                                         field='*',
                                         value='*')

edit_rich_schema_cred_def = AuthActionEdit(txn_type=RICH_SCHEMA_CRED_DEF,
                                           field='*',
                                           old_value='*',
                                           new_value='*')

add_rich_schema_pres_def = AuthActionAdd(txn_type=RICH_SCHEMA_PRES_DEF,
                                         field='*',
                                         value='*')

edit_rich_schema_pres_def = AuthActionEdit(txn_type=RICH_SCHEMA_PRES_DEF,
                                           field='*',
                                           old_value='*',
                                           new_value='*')

add_claim_def = AuthActionAdd(txn_type=CLAIM_DEF,
                              field='*',
                              value='*')

edit_claim_def = AuthActionEdit(txn_type=CLAIM_DEF,
                                field='*',
                                old_value='*',
                                new_value='*')

adding_new_node = AuthActionAdd(txn_type=NODE,
                                field='services',
                                value=['VALIDATOR'])

adding_new_node_with_empty_services = AuthActionAdd(txn_type=NODE,
                                                    field='services',
                                                    value=[])

demote_node = AuthActionEdit(txn_type=NODE,
                             field='services',
                             old_value=['VALIDATOR'],
                             new_value=[])

promote_node = AuthActionEdit(txn_type=NODE,
                              field='services',
                              old_value=[],
                              new_value=['VALIDATOR'])

change_node_ip = AuthActionEdit(txn_type=NODE,
                                field='node_ip',
                                old_value='*',
                                new_value='*')

change_node_port = AuthActionEdit(txn_type=NODE,
                                  field='node_port',
                                  old_value='*',
                                  new_value='*')

change_client_ip = AuthActionEdit(txn_type=NODE,
                                  field='client_ip',
                                  old_value='*',
                                  new_value='*')

change_client_port = AuthActionEdit(txn_type=NODE,
                                    field='client_port',
                                    old_value='*',
                                    new_value='*')

change_bls_key = AuthActionEdit(txn_type=NODE,
                                field='blskey',
                                old_value='*',
                                new_value='*')

start_upgrade = AuthActionAdd(txn_type=POOL_UPGRADE,
                              field='action',
                              value='start')

cancel_upgrade = AuthActionEdit(txn_type=POOL_UPGRADE,
                                field='action',
                                old_value='start',
                                new_value='cancel')

pool_restart = AuthActionAdd(txn_type=POOL_RESTART,
                             field='action',
                             value='*')

pool_config = AuthActionEdit(txn_type=POOL_CONFIG,
                             field='action',
                             old_value='*',
                             new_value='*')

auth_rule = AuthActionEdit(txn_type=AUTH_RULE,
                           field='*',
                           old_value='*',
                           new_value='*')

auth_rules = AuthActionEdit(txn_type=AUTH_RULES,
                            field='*',
                            old_value='*',
                            new_value='*')

validator_info = AuthActionAdd(txn_type=VALIDATOR_INFO,
                               field='*',
                               value='*')

add_revoc_reg_def = AuthActionAdd(txn_type=REVOC_REG_DEF,
                                  field='*',
                                  value='*')

add_revoc_reg_entry = AuthActionAdd(txn_type=REVOC_REG_ENTRY,
                                    field='*',
                                    value='*')

edit_revoc_reg_def = AuthActionEdit(txn_type=REVOC_REG_DEF,
                                    field='*',
                                    old_value='*',
                                    new_value='*')

edit_revoc_reg_entry = AuthActionEdit(txn_type=REVOC_REG_ENTRY,
                                      field='*',
                                      old_value='*',
                                      new_value='*')

edit_frozen_ledgers = AuthActionEdit(txn_type=LEDGERS_FREEZE,
                                     field='*',
                                     old_value='*',
                                     new_value='*')

# Anyone constraint
anyone_constraint = AuthConstraint(role='*',
                                   sig_count=1)

# No one constraint
no_one_constraint = AuthConstraintForbidden()

# Owner constraint
owner_constraint = AuthConstraint(role='*',
                                  sig_count=1,
                                  need_to_be_owner=True)

# Steward owner constraint
steward_owner_constraint = AuthConstraint(STEWARD, 1, need_to_be_owner=True)

# One Trustee constraint
one_trustee_constraint = AuthConstraint(TRUSTEE, 1)

# Three Trustee constraint
three_trustee_constraint = AuthConstraint(TRUSTEE, 3)

# Steward or Trustee constraint
steward_or_trustee_constraint = AuthConstraintOr([AuthConstraint(STEWARD, 1),
                                                  AuthConstraint(TRUSTEE, 1)])

# Endorser, Steward or Trustee constraint
endorser_or_steward_or_trustee_constraint = AuthConstraintOr([AuthConstraint(TRUSTEE, 1),
                                                              AuthConstraint(STEWARD, 1),
                                                              AuthConstraint(ENDORSER, 1)])
# Trustee or owner steward
trustee_or_owner_steward = AuthConstraintOr([AuthConstraint(TRUSTEE, 1),
                                             AuthConstraint(STEWARD, 1, need_to_be_owner=True)])

endorser_or_steward_or_trustee_owner_constraint = AuthConstraintOr(
    [AuthConstraint(TRUSTEE, 1, need_to_be_owner=True),
     AuthConstraint(STEWARD, 1, need_to_be_owner=True),
     AuthConstraint(ENDORSER, 1, need_to_be_owner=True)])

auth_map = OrderedDict([
    (add_new_trustee.get_action_id(), one_trustee_constraint),
    (add_new_steward.get_action_id(), one_trustee_constraint),
    (add_new_endorser.get_action_id(), steward_or_trustee_constraint),
    (add_new_network_monitor.get_action_id(), steward_or_trustee_constraint),
    (add_new_identity_owner.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (key_rotation.get_action_id(), owner_constraint),
    (txn_author_agreement.get_action_id(), one_trustee_constraint),
    (edit_txn_author_agreement.get_action_id(), one_trustee_constraint),
    (disable_txn_author_agreement.get_action_id(), one_trustee_constraint),
    (txn_author_agreement_aml.get_action_id(), one_trustee_constraint),
    (add_attrib.get_action_id(), owner_constraint),
    (edit_attrib.get_action_id(), owner_constraint),
    (add_schema.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_schema.get_action_id(), no_one_constraint),

    (add_json_ld_context.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_json_ld_context.get_action_id(), no_one_constraint),
    (add_rich_schema.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_rich_schema.get_action_id(), no_one_constraint),
    (add_rich_schema_encoding.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_rich_schema_encoding.get_action_id(), no_one_constraint),
    (add_rich_schema_mapping.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_rich_schema_mapping.get_action_id(), no_one_constraint),
    (add_rich_schema_cred_def.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_rich_schema_cred_def.get_action_id(), no_one_constraint),
    (add_rich_schema_pres_def.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_rich_schema_pres_def.get_action_id(), no_one_constraint),

    (add_claim_def.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (edit_claim_def.get_action_id(), owner_constraint),
    (adding_new_node.get_action_id(), steward_owner_constraint),
    (adding_new_node_with_empty_services.get_action_id(), steward_owner_constraint),
    (demote_node.get_action_id(), trustee_or_owner_steward),
    (promote_node.get_action_id(), trustee_or_owner_steward),
    (change_node_ip.get_action_id(), steward_owner_constraint),
    (change_node_port.get_action_id(), steward_owner_constraint),
    (change_client_ip.get_action_id(), steward_owner_constraint),
    (change_client_port.get_action_id(), steward_owner_constraint),
    (change_bls_key.get_action_id(), steward_owner_constraint),
    (start_upgrade.get_action_id(), one_trustee_constraint),
    (cancel_upgrade.get_action_id(), one_trustee_constraint),
    (pool_restart.get_action_id(), one_trustee_constraint),
    (pool_config.get_action_id(), one_trustee_constraint),
    (auth_rule.get_action_id(), one_trustee_constraint),
    (auth_rules.get_action_id(), one_trustee_constraint),
    (validator_info.get_action_id(), AuthConstraintOr([AuthConstraint(TRUSTEE, 1),
                                                       AuthConstraint(STEWARD, 1),
                                                       AuthConstraint(NETWORK_MONITOR, 1)])),
    (add_revoc_reg_def.get_action_id(), endorser_or_steward_or_trustee_constraint),
    (add_revoc_reg_entry.get_action_id(), endorser_or_steward_or_trustee_owner_constraint),
    (edit_revoc_reg_def.get_action_id(), owner_constraint),
    (edit_revoc_reg_entry.get_action_id(), owner_constraint),
    (edit_frozen_ledgers.get_action_id(), three_trustee_constraint),
])

# Edit Trustee:
auth_map_trustee = OrderedDict([
    (edit_role_actions[TRUSTEE][TRUSTEE].get_action_id(), owner_constraint),
    (edit_role_actions[TRUSTEE][STEWARD].get_action_id(), one_trustee_constraint),
    (edit_role_actions[TRUSTEE][ENDORSER].get_action_id(), one_trustee_constraint),
    (edit_role_actions[TRUSTEE][NETWORK_MONITOR].get_action_id(), one_trustee_constraint),
    (edit_role_actions[TRUSTEE][IDENTITY_OWNER].get_action_id(), one_trustee_constraint),
])
auth_map.update(auth_map_trustee)

# Edit Steward
auth_map_steward = OrderedDict([
    (edit_role_actions[STEWARD][TRUSTEE].get_action_id(), one_trustee_constraint),
    (edit_role_actions[STEWARD][STEWARD].get_action_id(), owner_constraint),
    (edit_role_actions[STEWARD][ENDORSER].get_action_id(), one_trustee_constraint),
    (edit_role_actions[STEWARD][NETWORK_MONITOR].get_action_id(), one_trustee_constraint),
    (edit_role_actions[STEWARD][IDENTITY_OWNER].get_action_id(), one_trustee_constraint),
])
auth_map.update(auth_map_steward)

# Edit Endorser
auth_map_endorser = OrderedDict([
    (edit_role_actions[ENDORSER][TRUSTEE].get_action_id(), one_trustee_constraint),
    (edit_role_actions[ENDORSER][STEWARD].get_action_id(), one_trustee_constraint),
    (edit_role_actions[ENDORSER][ENDORSER].get_action_id(), owner_constraint),
    (edit_role_actions[ENDORSER][NETWORK_MONITOR].get_action_id(), one_trustee_constraint),
    (edit_role_actions[ENDORSER][IDENTITY_OWNER].get_action_id(), one_trustee_constraint),
])
auth_map.update(auth_map_endorser)

# Edit Network Monitor
auth_map_network_monitor = OrderedDict([
    (edit_role_actions[NETWORK_MONITOR][TRUSTEE].get_action_id(), one_trustee_constraint),
    (edit_role_actions[NETWORK_MONITOR][STEWARD].get_action_id(), one_trustee_constraint),
    (edit_role_actions[NETWORK_MONITOR][ENDORSER].get_action_id(), steward_or_trustee_constraint),
    (edit_role_actions[NETWORK_MONITOR][NETWORK_MONITOR].get_action_id(), owner_constraint),
    (edit_role_actions[NETWORK_MONITOR][IDENTITY_OWNER].get_action_id(), steward_or_trustee_constraint),
])
auth_map.update(auth_map_network_monitor)

# Edit Identity Owner
auth_map_identity_owner = OrderedDict([
    (edit_role_actions[IDENTITY_OWNER][TRUSTEE].get_action_id(), one_trustee_constraint),
    (edit_role_actions[IDENTITY_OWNER][STEWARD].get_action_id(), one_trustee_constraint),
    (edit_role_actions[IDENTITY_OWNER][ENDORSER].get_action_id(), steward_or_trustee_constraint),
    (edit_role_actions[IDENTITY_OWNER][NETWORK_MONITOR].get_action_id(), steward_or_trustee_constraint),
    (edit_role_actions[IDENTITY_OWNER][IDENTITY_OWNER].get_action_id(), owner_constraint),
])
auth_map.update(auth_map_identity_owner)



import pytest

from indy_node.test.state_proof.helper import sdk_submit_operation_and_get_result
from plenum.common.constants import TARGET_NYM, TXN_TYPE, RAW
from indy_common.constants import GET_ATTR

# fixtures
from indy_node.test.attrib_txn.test_nym_attrib import attributeData, \
    attributeName, attributeValue, sdk_added_raw_attribute


def test_client_gets_read_reply_from_1_node_only(looper,
                                                 nodeSetWithOneNodeResponding,
                                                 sdk_added_raw_attribute,
                                                 attributeName,
                                                 sdk_pool_handle,
                                                 sdk_wallet_trustee):
    """
    Tests that client could send get-requests to only one node instead of n
    """
    # Prepare and send get-request
    get_attr_operation = {
        TARGET_NYM: sdk_added_raw_attribute['result']['txn']['data']['dest'],
        TXN_TYPE: GET_ATTR,
        RAW: attributeName
    }

    sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                        sdk_wallet_trustee,
                                        get_attr_operation)



import pytest

from indy_common.constants import ENDORSER
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_pres_def_handler import \
    RichSchemaPresDefHandler
from indy_node.test.request_handlers.helper import add_to_idr
from indy_node.test.request_handlers.rich_schema.helper import rich_schema_pres_def_request
from plenum.common.constants import TRUSTEE


@pytest.fixture()
def pres_def_handler(db_manager, write_auth_req_validator):
    return RichSchemaPresDefHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def pres_def_req(pres_def_handler):
    req = rich_schema_pres_def_request()
    add_to_idr(pres_def_handler.database_manager.idr_cache, req.identifier, TRUSTEE)
    add_to_idr(pres_def_handler.database_manager.idr_cache, req.endorser, ENDORSER)
    return req


def test_schema_dynamic_validation_passes(pres_def_handler, pres_def_req):
    pres_def_handler.dynamic_validation(pres_def_req, 0)



from abc import abstractmethod

from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler
from plenum.common.exceptions import InvalidClientRequest
from indy_common.config_util import getConfig
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager


class AbstractRichSchemaReadRequestHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager, txn_type, ledger_id):
        super().__init__(database_manager, txn_type, ledger_id)
        self.config = getConfig()

    def _enabled(self) -> bool:
        return self.config.ENABLE_RICH_SCHEMAS

    def _validate_enabled(self, request: Request):
        if not self._enabled():
            raise InvalidClientRequest(request.identifier, request.reqId, "RichSchema queries are disabled")

    @abstractmethod
    def get_result(self, request: Request):
        super().get_result(request)
        self._validate_enabled(request)



import json
import pytest

from indy_node.test.helper import sdk_add_attribute_and_check
from indy_common.auth import Authoriser
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import sdk_sign_request_from_dict, sdk_send_and_check, sdk_send_signed_requests, \
    sdk_get_bad_response, sdk_sign_and_submit_req, sdk_get_and_check_replies

from indy_node.test.anon_creds.conftest import claim_def, build_revoc_reg_entry_for_given_revoc_reg_def
from indy_node.test.attrib_txn.test_nym_attrib import attributeData, attributeName, attributeValue
from indy_node.test.anon_creds.conftest import build_revoc_def_by_default
from indy_node.test.schema.test_send_get_schema import send_schema_seq_no
from indy_node.test.schema.test_send_get_schema import send_schema_req

from indy.anoncreds import issuer_create_schema
from indy.ledger import build_schema_request


def test_client_cant_send_nym(looper,
                              txnPoolNodeSet,
                              sdk_wallet_client,
                              sdk_wallet_endorser,
                              sdk_pool_handle, attributeData):
    # Endorser can create schema
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_endorser)

    # client cant create another client NYM
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_client)
    e.match('Rule for this action is')


def test_client_can_send_attrib(looper,
                                sdk_wallet_client,
                                sdk_wallet_endorser,
                                sdk_pool_handle,
                                attributeData):
    _, client_did = sdk_wallet_client

    # Client as an owner can add attribute to his NYM
    sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_client,
                                attributeData, client_did)

    # another client or endorser cannot add attribute to another NYM
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_endorser,
                                    attributeData, client_did)
    assert e.match('can not touch raw field since only the owner can modify it')


def test_client_cant_send_schema(looper,
                                 txnPoolNodeSet,
                                 sdk_wallet_client,
                                 sdk_wallet_endorser,
                                 sdk_pool_handle):
    # Endorser can create schema
    _, identifier = sdk_wallet_endorser
    _, schema_json = looper.loop.run_until_complete(
        issuer_create_schema(identifier, "another_name", "2.0", json.dumps(["first", "last"])))
    request = looper.loop.run_until_complete(build_schema_request(identifier, schema_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_endorser, request)])

    # Client cant create schema
    _, identifier = sdk_wallet_client
    _, schema_json = looper.loop.run_until_complete(
        issuer_create_schema(identifier, "another_name", "2.0", json.dumps(["first", "last"])))
    request = looper.loop.run_until_complete(build_schema_request(identifier, schema_json))
    sdk_get_bad_response(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, request)],
                         RequestRejectedException, 'Rule for this action is')


def test_client_cant_send_claim_def(looper,
                                    txnPoolNodeSet,
                                    sdk_wallet_client,
                                    sdk_wallet_endorser,
                                    sdk_pool_handle,
                                    claim_def):
    # Endorser can create claim_def
    req = sdk_sign_request_from_dict(looper, sdk_wallet_endorser, claim_def)
    sdk_send_and_check([json.dumps(req)], looper, txnPoolNodeSet, sdk_pool_handle)

    # Client cant send create
    req = sdk_sign_request_from_dict(looper, sdk_wallet_client, claim_def)
    req = sdk_send_signed_requests(sdk_pool_handle, [json.dumps(req)])
    sdk_get_bad_response(looper, req, RequestRejectedException, 'Rule for this action is')



import pytest
from plenum.common.constants import CURRENT_PROTOCOL_VERSION

from indy_common.req_utils import get_write_claim_def_signature_type, \
    get_write_claim_def_schema_ref, get_write_claim_def_tag, get_write_claim_def_public_keys, \
    get_read_claim_def_signature_type, get_read_claim_def_schema_ref, get_read_claim_def_tag, get_read_claim_def_from, \
    get_txn_claim_def_public_keys, get_txn_claim_def_tag, get_txn_claim_def_schema_ref, get_txn_claim_def_signature_type
from indy_common.types import SafeRequest
from plenum.common.txn_util import reqToTxn


@pytest.fixture()
def write_claim_def_request(request):
    req = {
        'operation': {
            'type': '102',
            'signature_type': 'CL1',
            'ref': 18,
            'tag': 'key111',
            'data': {
                'primary': {'primaryKey1': 'a'},
                'revocation': {'revocationKey1': 'b'}
            },
        },

        'identifier': 'L5AD5g65TDQr1PPHHRoiGf',
        'reqId': 1514280215504647,
        'protocolVersion': CURRENT_PROTOCOL_VERSION,
        'signature': '5ZTp9g4SP6t73rH2s8zgmtqdXyTuSMWwkLvfV1FD6ddHCpwTY5SAsp8YmLWnTgDnPXfJue3vJBWjy89bSHvyMSdS'
    }
    return SafeRequest(**req)


@pytest.fixture()
def write_claim_def_request_no_signature_type(write_claim_def_request):
    del write_claim_def_request.operation['signature_type']
    return write_claim_def_request


@pytest.fixture()
def write_claim_def_request_no_tag(write_claim_def_request):
    del write_claim_def_request.operation['tag']
    return write_claim_def_request


@pytest.fixture(scope="module")
def read_claim_def_request():
    req = {
        'operation': {
            'type': '108',
            'signature_type': 'CL1',
            'ref': 18,
            'tag': 'key111',
            'origin': 'L5AD5g65TDQr1PPHHRoiGf'
        },

        'identifier': 'E5AD5g65TDQr1PPHHRoiGf',
        'reqId': 1514280215504647,
        'protocolVersion': CURRENT_PROTOCOL_VERSION
    }
    return SafeRequest(**req)


@pytest.fixture(scope="module")
def read_claim_def_request_no_signature_type(read_claim_def_request):
    del read_claim_def_request.operation['signature_type']
    return read_claim_def_request


@pytest.fixture(scope="module")
def read_claim_def_request_no_tag(read_claim_def_request):
    del read_claim_def_request.operation['tag']
    return read_claim_def_request


def test_get_write_claim_def_signature_type(write_claim_def_request):
    assert 'CL1' == get_write_claim_def_signature_type(write_claim_def_request)


def test_get_write_claim_def_signature_type_default(write_claim_def_request_no_signature_type):
    assert 'CL' == get_write_claim_def_signature_type(write_claim_def_request_no_signature_type)


def test_get_write_claim_def_schema_ref(write_claim_def_request):
    assert 18 == get_write_claim_def_schema_ref(write_claim_def_request)


def test_get_write_claim_def_tag(write_claim_def_request):
    assert 'key111' == get_write_claim_def_tag(write_claim_def_request)


def test_get_write_claim_def_tag_default(write_claim_def_request_no_tag):
    assert 'tag' == get_write_claim_def_tag(write_claim_def_request_no_tag)


def test_get_write_claim_public_keys(write_claim_def_request):
    assert {'primary': {'primaryKey1': 'a'}, 'revocation': {'revocationKey1': 'b'}} == \
        get_write_claim_def_public_keys(write_claim_def_request)


def test_get_txn_claim_def_signature_type(write_claim_def_request):
    txn = reqToTxn(write_claim_def_request)
    assert 'CL1' == get_txn_claim_def_signature_type(txn)


def test_get_txn_claim_def_signature_type_default(write_claim_def_request_no_signature_type):
    txn = reqToTxn(write_claim_def_request_no_signature_type)
    assert 'CL' == get_txn_claim_def_signature_type(txn)


def test_get_txn_claim_def_schema_ref(write_claim_def_request):
    txn = reqToTxn(write_claim_def_request)
    assert 18 == get_txn_claim_def_schema_ref(txn)


def test_get_txn_claim_def_tag(write_claim_def_request):
    txn = reqToTxn(write_claim_def_request)
    assert 'key111' == get_txn_claim_def_tag(txn)


def test_get_txn_claim_def_tag_default(write_claim_def_request_no_tag):
    txn = reqToTxn(write_claim_def_request_no_tag)
    assert 'tag' == get_txn_claim_def_tag(txn)


def test_get_txn_claim_public_keys(write_claim_def_request):
    txn = reqToTxn(write_claim_def_request)
    assert {'primary': {'primaryKey1': 'a'}, 'revocation': {'revocationKey1': 'b'}} == \
        get_txn_claim_def_public_keys(txn)


def test_get_read_claim_def_signature_type(read_claim_def_request):
    assert 'CL1' == get_read_claim_def_signature_type(read_claim_def_request)


def test_get_read_claim_def_signature_type_default(read_claim_def_request_no_signature_type):
    assert 'CL' == get_read_claim_def_signature_type(read_claim_def_request_no_signature_type)


def test_get_read_claim_def_schema_ref(read_claim_def_request):
    assert 18 == get_read_claim_def_schema_ref(read_claim_def_request)


def test_get_read_claim_def_tag(read_claim_def_request):
    assert 'key111' == get_read_claim_def_tag(read_claim_def_request)


def test_get_read_claim_def_tag_default(read_claim_def_request_no_tag):
    assert 'tag' == get_read_claim_def_tag(read_claim_def_request_no_tag)


def test_get_read_claim_def_from(read_claim_def_request):
    assert 'L5AD5g65TDQr1PPHHRoiGf' == get_read_claim_def_from(read_claim_def_request)



import multiprocessing
import os
import shutil

import pytest

from stp_core.loop.eventually import eventually
from indy_common.version import src_version_cls
from indy_node.server.upgrader import Upgrader
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion

from indy_node.test.upgrade.helper import (
    NodeControlToolExecutor as NCT,
    sendUpgradeMessage,
    nodeControlGeneralMonkeypatching,
    bumpedVersion
)

m = multiprocessing.Manager()
# TODO why do we expect that
whitelist = ['Unexpected error in _upgrade test']


@pytest.fixture(scope='module')
def tconf(tconf, tdir):
    tconf.LOG_DIR = tdir
    yield tconf


def testNodeControlCreatesBackups(monkeypatch, tdir, looper, tconf):
    version = bumpedVersion()
    stdout = 'teststdout'
    curr_src_ver = Upgrader.get_src_version()

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        monkeypatch.setattr(tool, '_remove_old_backups', lambda *x: None)

    def checkBackup(tool):
        assert os.path.isfile('{}.{}'.format(
            tool._backup_name(curr_src_ver.release), tool.backup_format))

    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: DebianVersion(
            version, upstream_cls=src_version_cls())
    )

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform, config=tconf)
    try:
        sendUpgradeMessage(version)
        looper.run(eventually(checkBackup, nct.tool))
    finally:
        clean_dir(nct.tool.base_dir)
        nct.stop()

def clean_dir(dir):
    for file in os.listdir(dir):
        file = os.path.join(dir, file)
        if os.path.isfile(file):
            os.remove(file)
        if os.path.isdir(file):
            shutil.rmtree(file, ignore_errors=True)



from indy_node.server.upgrade_log import UpgradeLog
from indy_node.test.upgrade.helper import count_action_log_package


def test_pool_upgrade_force_scheduled_only_once(validUpgradeExpForceTrue, upgradeScheduledExpForceTrue, nodeSet):
    for node in nodeSet:
        assert count_action_log_package(list(node.upgrader._actionLog), validUpgradeExpForceTrue['package']) == 1
        assert node.upgrader._actionLog.last_event.ev_type == UpgradeLog.Events.scheduled



import copy
import json

import pytest

from indy_common.authorize import auth_map
from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import IDENTITY_OWNER, AuthConstraint
from indy_common.constants import REVOC_REG_ENTRY, PREV_ACCUM, VALUE, CRED_DEF_ID, REVOC_TYPE, TAG, REVOC_REG_DEF_ID, \
    ACCUM, ISSUED, REVOKED
from indy_common.state import domain
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.auth_rule.auth_framework.revoc_reg_def import AddRevocRegDefTest
from plenum.common.constants import TXN_TYPE
from plenum.common.exceptions import RequestRejectedException
from plenum.common.types import f, OPERATION
from plenum.common.util import randomString
from plenum.test.helper import sdk_sign_request_from_dict, sdk_send_and_check
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.helper import build_auth_rule_request_json


class AddRevocRegEntryTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.claim_def_req = None
        self.revoc_reg_def_req = None
        self.first_entry = None
        self.next_reg_entry = None

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        rrd_test = AddRevocRegDefTest(self.env, auth_map.add_revoc_reg_def.get_action_id())
        rrd_test.prepare()
        self.claim_def_req = rrd_test.claim_def_req
        self.first_revoc_reg_def_req = rrd_test.send_revoc_reg_def(self.claim_def_req, wallet=self.trustee_wallet)[0][0]
        self.second_revoc_reg_def_req = rrd_test.send_revoc_reg_def(self.claim_def_req, wallet=self.trustee_wallet)[0][0]

        self.first_entry = self.build_reg_entry_req(revoc_reg_req=self.first_revoc_reg_def_req,
                                                    revoked=[1, 2, 3])
        self.next_reg_entry = self.build_reg_entry_req(revoc_reg_req=self.second_revoc_reg_def_req,
                                                       revoked=[4, 5, 6])

    def build_revoc_reg_entry_for_given_revoc_reg_def(self,
                                                      revoc_def_req):
        path = ":".join([revoc_def_req[f.IDENTIFIER.nm],
                         domain.MARKER_REVOC_DEF,
                         revoc_def_req[OPERATION][CRED_DEF_ID],
                         revoc_def_req[OPERATION][REVOC_TYPE],
                         revoc_def_req[OPERATION][TAG]])
        data = {
            REVOC_REG_DEF_ID: path,
            TXN_TYPE: REVOC_REG_ENTRY,
            REVOC_TYPE: revoc_def_req[OPERATION][REVOC_TYPE],
            VALUE: {
                PREV_ACCUM: randomString(10),
                ACCUM: randomString(10),
                ISSUED: [],
                REVOKED: [],
            }
        }
        return data

    def build_reg_entry_req(self, revoc_reg_req, revoked):
        data = self.build_revoc_reg_entry_for_given_revoc_reg_def(revoc_reg_req)
        data[VALUE][REVOKED] = revoked
        del data[VALUE][PREV_ACCUM]
        return data

    def run(self):
        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            self.send_revoc_reg_entry(data=self.first_entry, wallet=self.trustee_wallet)

        # Step 3. Check, that new auth rule is used
        self.send_revoc_reg_entry(data=self.first_entry, wallet=self.new_default_wallet)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.send_revoc_reg_entry(data=self.next_reg_entry, wallet=self.trustee_wallet)

    def result(self):
        pass

    def send_revoc_reg_entry(self, data, wallet=None):
        wallet = wallet if wallet else self.trustee_wallet
        req = self.build_txn_for_revoc_def_entry_by_default(wallet,
                                                            data)
        return sdk_send_and_check([json.dumps(req)], self.looper, self.env.txnPoolNodeSet, self.sdk_pool_handle)

    def build_txn_for_revoc_def_entry_by_default(self,
                                                 wallet,
                                                 data):
        req = sdk_sign_request_from_dict(self.looper, wallet, data)
        return req

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=REVOC_REG_ENTRY,
            field='*',
            new_value='*',
            constraint=constraint.as_dict
        )


class EditRevocRegEntryTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.rre_test = None
        self.claim_def_req = None
        self.first_edit = None
        self.second_edit = None

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        self.rre_test = AddRevocRegEntryTest(self.env, auth_map.edit_revoc_reg_entry.get_action_id())
        self.rre_test.prepare()
        self.claim_def_req = self.rre_test.claim_def_req
        reg_entry = self.rre_test.first_entry

        self.first_edit = copy.deepcopy(reg_entry)
        self.second_edit = copy.deepcopy(reg_entry)
        self.rre_test.send_revoc_reg_entry(reg_entry, wallet=self.trustee_wallet)
        self.first_edit[VALUE][REVOKED] = [10, 20]
        self.first_edit[VALUE][PREV_ACCUM] = reg_entry[VALUE][ACCUM]
        self.first_edit[VALUE][ACCUM] = randomString(10)

        self.second_edit[VALUE][ACCUM] = randomString(10)
        self.second_edit[VALUE][REVOKED] = [30, 40]
        self.second_edit[VALUE][PREV_ACCUM] = self.first_edit[VALUE][ACCUM]

    def run(self):
        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            self.rre_test.send_revoc_reg_entry(data=self.first_edit, wallet=self.trustee_wallet)

        # Step 3. Check, that new auth rule is used
        self.rre_test.send_revoc_reg_entry(data=self.first_edit, wallet=self.new_default_wallet)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.rre_test.send_revoc_reg_entry(data=self.second_edit, wallet=self.trustee_wallet)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=EDIT_PREFIX,
            auth_type=REVOC_REG_ENTRY,
            field='*',
            old_value='*',
            new_value='*',
            constraint=constraint.as_dict
        )
